from typing import Optional
from typing import Tuple
from typing import List
from typing import Iterator

class BaseConverter(object):
    IGNORE = "converter_ignore"

    def __init__(self):
        self.format_name = None
        self.output_dir = None
        self.run_name = None
        self.format_name = None
        self.module_name: Optional[str] = None
        self.version = None
        self.conf: dict = {}
        self.input_path = ""
        self.input_paths: Optional[List[str]] = None

    def check_format(self, *__args__, **__kwargs__):
        pass

    def setup(self, *__args__, **__kwargs__):
        pass

    def convert_line(self, *__args__, **__kwargs__) -> List[dict]:
        return []

    def convert_file(self, file, *__args__, exc_handler=None, **__kwargs__) -> Iterator[Tuple[int, List[dict]]]:
        line_no = 0
        for line in file:
            line_no += 1
            try:
                yield line_no, self.convert_line(line)
            except Exception as e:
                if exc_handler:
                    exc_handler(line_no, e)
                else:
                    raise e
        return None

    def addl_operation_for_unique_variant(self, __wdict__, __wdict_no__):
        pass



complementary_base = {
    "A": "T",
    "T": "A",
    "C": "G",
    "G": "C",
    "-": "-",
    "": "",
    "N": "N",
}

aa_123 = {
    "A": "Ala",
    "C": "Cys",
    "E": "Glu",
    "D": "Asp",
    "G": "Gly",
    "F": "Phe",
    "I": "Ile",
    "H": "His",
    "K": "Lys",
    "M": "Met",
    "L": "Leu",
    "N": "Asn",
    "Q": "Gln",
    "P": "Pro",
    "S": "Ser",
    "R": "Arg",
    "T": "Thr",
    "W": "Trp",
    "V": "Val",
    "Y": "Tyr",
    "*": "Ter",
    "": "",
}

aa_321 = {
    "Asp": "D",
    "Ser": "S",
    "Gln": "Q",
    "Lys": "K",
    "Trp": "W",
    "Asn": "N",
    "Pro": "P",
    "Thr": "T",
    "Phe": "F",
    "Ala": "A",
    "Gly": "G",
    "Cys": "C",
    "Ile": "I",
    "Leu": "L",
    "His": "H",
    "Arg": "R",
    "Met": "M",
    "Val": "V",
    "Glu": "E",
    "Tyr": "Y",
    "Ter": "*",
    "": "",
}

codon_table = {
    "ATG": "M",
    "GCT": "A",
    "GCC": "A",
    "GCA": "A",
    "GCG": "A",
    "TGT": "C",
    "TGC": "C",
    "GAT": "D",
    "GAC": "D",
    "GAA": "E",
    "GAG": "E",
    "TTT": "F",
    "TTC": "F",
    "GGT": "G",
    "GGC": "G",
    "GGA": "G",
    "GGG": "G",
    "CAT": "H",
    "CAC": "H",
    "ATT": "I",
    "ATC": "I",
    "ATA": "I",
    "AAA": "K",
    "AAG": "K",
    "TTA": "L",
    "TTG": "L",
    "CTT": "L",
    "CTC": "L",
    "CTA": "L",
    "CTG": "L",
    "AAT": "N",
    "AAC": "N",
    "CCT": "P",
    "CCC": "P",
    "CCA": "P",
    "CCG": "P",
    "CAA": "Q",
    "CAG": "Q",
    "TCT": "S",
    "TCC": "S",
    "TCA": "S",
    "TCG": "S",
    "AGT": "S",
    "AGC": "S",
    "ACT": "T",
    "ACC": "T",
    "ACA": "T",
    "ACG": "T",
    "CGT": "R",
    "CGC": "R",
    "CGA": "R",
    "CGG": "R",
    "AGA": "R",
    "AGG": "R",
    "GTT": "V",
    "GTC": "V",
    "GTA": "V",
    "GTG": "V",
    "TGG": "W",
    "TAT": "Y",
    "TAC": "Y",
    "TGA": "*",
    "TAA": "*",
    "TAG": "*",
    "AUG": "M",
    "GCU": "A",
    "GCC": "A",
    "GCA": "A",
    "GCG": "A",
    "UGU": "C",
    "UGC": "C",
    "GAU": "D",
    "GAC": "D",
    "GAA": "E",
    "GAG": "E",
    "UUU": "F",
    "UUC": "F",
    "GGU": "G",
    "GGC": "G",
    "GGA": "G",
    "GGG": "G",
    "CAU": "H",
    "CAC": "H",
    "AUU": "I",
    "AUC": "I",
    "AUA": "I",
    "AAA": "K",
    "AAG": "K",
    "UUA": "L",
    "UUG": "L",
    "CUU": "L",
    "CUC": "L",
    "CUA": "L",
    "CUG": "L",
    "AAU": "N",
    "AAC": "N",
    "CCU": "P",
    "CCC": "P",
    "CCA": "P",
    "CCG": "P",
    "CAA": "Q",
    "CAG": "Q",
    "UCU": "S",
    "UCC": "S",
    "UCA": "S",
    "UCG": "S",
    "AGU": "S",
    "AGC": "S",
    "ACU": "T",
    "ACC": "T",
    "ACA": "T",
    "ACG": "T",
    "CGU": "R",
    "CGC": "R",
    "CGA": "R",
    "CGG": "R",
    "AGA": "R",
    "AGG": "R",
    "GUU": "V",
    "GUC": "V",
    "GUA": "V",
    "GUG": "V",
    "UGG": "W",
    "UAU": "Y",
    "UAC": "Y",
    "UGA": "*",
    "UAA": "*",
    "UAG": "*",
}


def trim_input_left_adjust(ref, alt, pos, strand):
    pos = int(pos)
    reflen = len(ref)
    altlen = len(alt)
    minlen = min(reflen, altlen)
    new_ref = ref
    new_alt = alt
    new_pos = pos
    for nt_pos in range(0, minlen):
        if ref[reflen - nt_pos - 1] == alt[altlen - nt_pos - 1]:
            new_ref = ref[: reflen - nt_pos - 1]
            new_alt = alt[: altlen - nt_pos - 1]
        else:
            break
    new_ref_len = len(new_ref)
    new_alt_len = len(new_alt)
    minlen = min(new_ref_len, new_alt_len)
    new_ref2 = new_ref
    new_alt2 = new_alt
    for nt_pos in range(0, minlen):
        if new_ref[nt_pos] == new_alt[nt_pos]:
            if strand == "+":
                new_pos += 1
            elif strand == "-":
                new_pos -= 1
            new_ref2 = new_ref[nt_pos + 1 :]
            new_alt2 = new_alt[nt_pos + 1 :]
        else:
            new_ref2 = new_ref[nt_pos:]
            new_alt2 = new_alt[nt_pos:]
            break
    return new_ref2, new_alt2, new_pos


def normalize_variant_left(strand, pos, ref, alt):
    reflen = len(ref)
    altlen = len(alt)
    # Returns without change if same single nucleotide for ref and alt.
    if reflen == 1 and altlen == 1 and ref == alt:
        return pos, ref, alt
    # Trimming from the start and then the end of the sequence
    # where the sequences overlap with the same nucleotides
    new_ref2, new_alt2, new_pos = trim_input_left_adjust(ref, alt, pos, strand)
    if new_ref2 == "" or new_ref2 == ".":
        new_ref2 = "-"
    if new_alt2 == "" or new_alt2 == ".":
        new_alt2 = "-"
    return new_pos, new_ref2, new_alt2


def normalize_variant_dict_left(wdict):
    from ..exceptions import NoVariantError

    chrom = wdict["chrom"]
    if not chrom.startswith("chr"):
        wdict["chrom"] = "chr" + chrom
    p, r, a = (
        int(wdict["pos"]),
        wdict["ref_base"],
        wdict["alt_base"],
    )
    (
        new_pos,
        new_ref,
        new_alt,
    ) = normalize_variant_left("+", p, r, a)
    wdict["pos"] = new_pos
    wdict["ref_base"] = new_ref
    wdict["alt_base"] = new_alt
    if wdict["ref_base"] == wdict["alt_base"]:
        raise NoVariantError()
    return wdict


def reverse_complement(bases):
    return "".join([complementary_base[base] for base in bases[::-1]])


def switch_strand(bases, start_strand=None, dest_strand=None, pos=0):
    rev_comp = reverse_complement(bases)
    if start_strand == "-" or dest_strand == "+":
        new_pos = pos + len(bases.replace("-", "")) - 1
    elif start_strand == "+" or dest_strand == "-":
        new_pos = pos - len(bases.replace("-", "")) + 1
    else:
        err_msg = "start_strand or dest_strand must be specified as + or -"
        raise ValueError(err_msg)
    return rev_comp, new_pos


def aa_let_to_abbv(lets):
    return "".join([aa_123[x] for x in lets])


def aa_abbv_to_let(abbvs):
    if type(abbvs) != str:
        raise TypeError("Expected str not %s" % type(abbvs).__name__)
    if len(abbvs) % 3 != 0:
        raise ValueError("Must be evenly divisible by 3")
    out = ""
    for i in range(0, len(abbvs), 3):
        abbv = abbvs[i].upper() + abbvs[i + 1 : i + 3].lower()
        out += aa_321[abbv]
    return out


def translate_codon(bases, fallback=None):
    if len(bases) != 3:
        if fallback is None:
            return KeyError(bases)
        else:
            return fallback
    else:
        return codon_table[bases]



class ExpectedException(Exception):
    halt = False
    handled = False
    traceback = True
    msg = ""
    returncode = 1

    def __init__(self, msg=""):
        self.msg = msg
        super().__init__(f"{msg}")


class NormalExit(ExpectedException):
    traceback = False
    returncode = 0


class NoGenomeException(ExpectedException):
    traceback = False
    halt = True

    def __init__(self):
        super().__init__("genome assembly should be selected.")


class InvalidGenomeAssembly(ExpectedException):
    traceback = False
    halt = True

    def __init__(self, genome_assembly):
        super().__init__(f"{genome_assembly} is an invalid genome assembly.")


class InvalidData(ExpectedException):
    nolog = True


class ConfigurationError(ExpectedException):
    traceback = True


class BadFormatError(InvalidData):
    pass


class IgnoredVariant(InvalidData):
    traceback = False

class NoAlternateAllele(ExpectedException):
    traceback = False

class LiftoverFailure(InvalidData):
    traceback = False
    pass


class FileIntegrityError(Exception):
    traceback = False

    def __init__(self, path):
        super().__init__(path)


class KillInstallException(ExpectedException):
    traceback = False

    pass


class InvalidFilter(ExpectedException):
    traceback = False

    def __init__(self, wrong_samples, wrong_colnames):
        self.msg = []
        if len(wrong_samples) > 0:
            self.msg.append(
                "Filter sample names do not exist: " + " ".join(wrong_samples)
            )
        if len(wrong_colnames) > 0:
            self.msg.append(
                "Filter column names do not exist: " + " ".join(wrong_colnames)
            )
        self.msg = "\n".join(self.msg)

    def __str__(self):
        return str(self.msg)


class ModuleInstallationError(ExpectedException):
    traceback = False
    halt = False

    def __init__(self, msg):
        super().__init__(msg)


class ModuleNotExist(ExpectedException):
    traceback = False
    halt = True
    returncode = 3

    def __init__(self, module_name, msg=None):
        if msg:
            super().__init__(msg)
        else:
            super().__init__(f"module [{module_name}] does not exist.")


class InvalidModule(ExpectedException):
    traceback = False

    def __init__(self, module_name):
        super().__init__(f"module {module_name} is invalid.")


class NoConverterFound(ExpectedException):
    traceback = False
    halt = True

    def __init__(self, input_file):
        super().__init__(f"No converter was found for {input_file}.")


class NoVariantError(ExpectedException):
    traceback = False

    def __init__(self):
        super().__init__("Reference and alternate alleles are the same.")


class NoInput(ExpectedException):
    traceback = False
    halt = True

    def __init__(self):
        import traceback
        super().__init__("No valid input was given.\n" + str(traceback.format_stack()))


class InvalidInputFormat(ExpectedException):
    traceback = False
    halt = True

    def __init__(self, fmt=""):
        super().__init__(f"invalid input format: {fmt}")


class SystemMissingException(ExpectedException):
    traceback = False

    def __init__(self, msg=""):
        if msg is not None and msg != "":
            msg = f"OakVar system component is missing ({msg}).\nPlease run 'ov system setup' to set up OakVar."
        else:
            msg = f"OakVar system component is missing.\nPlease run 'ov system setup' to set up OakVar."
        super().__init__(msg)


class NoModulesDir(ExpectedException):
    traceback = False

    def __init__(self):
        super().__init__("no modules directory. Run `ov system setup` to setup.")


class NoSystemModule(ExpectedException):
    traceback = False

    def __init__(self):
        super().__init__("no system module. Run `ov system setup` to setup.")


class IncompatibleResult(ExpectedException):
    traceback = False

    def __init__(self):
        super().__init__(
            "incompatible result file version. Please report with `ov issue`."
        )


class ModuleLoadingError(ExpectedException):
    traceback = True
    halt = True

    def __init__(self, module_name):
        super().__init__(msg=f"module loading error for {module_name}")


class UnknownInputFormat(ExpectedException):
    traceback = False
    halt = True

    def __init__(self, input_format):
        super().__init__(
            f"converter for {input_format} is not found. Please check if a converter is available for the format with `ov module ls -a -t converter`."
        )


class AbsentJobConf(ExpectedException):
    traceback = False
    halt = True

    def __init__(self, job_conf_path):
        super().__init__(f"{job_conf_path} does not exist.")


class StoreIncorrectLogin(ExpectedException):
    traceback = False
    halt = True

    def __init__(self):
        super().__init__(f"store login is incorrect.")


class StoreServerError(ExpectedException):
    traceback = False
    halt = True

    def __init__(self, status_code=500, text=None):
        if text is None:
            super().__init__(f"store server error [{status_code}].")
        else:
            super().__init__(f"store server error [{status_code}]: {text}.")


class InternetConnectionError(ExpectedException):
    def __init__(self):
        super().__init__(f"internet connection is unavailable.")


class ModuleVersionError(ExpectedException):
    def __init__(self, module_name, version, msg=None):
        if msg:
            super().__init__(msg)
        else:
            super().__init__(f"{module_name}=={version} does not exist.")


class ModuleToSkipInstallation(ExpectedException):
    traceback = False
    def __init__(self, module_name, msg=None):
        if msg:
            super().__init__(f"{msg}")
        else:
            super().__init__(f"Skipping installation of {module_name}")


class SetupError(ExpectedException):
    traceback = True
    halt = True

    def __init__(self, module_name=None, msg=None):
        if msg:
            super().__init__(msg=msg)
        else:
            if module_name:
                super().__init__(msg=f"setup for {module_name}")
            else:
                super().__init__(msg=f"setup error")


class LoggerError(ExpectedException):
    traceback = True
    halt = True

    def __init__(self, module_name=None):
        if module_name is not None:
            super().__init__(f"logger is None for {module_name}")
        else:
            super().__init__(f"logger is None")


class IncompleteModuleError(ExpectedException):
    def __init__(self, module_name):
        super().__init__(f"incomplete module {module_name}")


class ResultMissingMandatoryColumnError(ExpectedException):
    traceback = False
    halt = False

    def __init__(self, dbpath, cols):
        super().__init__(f"Error: {dbpath} lacks {cols}")


class FilterLoadingError(ExpectedException):
    def __init__(self):
        super().__init__(f"filter loading error")


class ParserError(ExpectedException):
    traceback = True
    halt = True

    def __init__(self, module_name=None):
        if module_name is not None:
            super().__init__(f"parser loading error for {module_name}")
        else:
            super().__init__(f"parser loading error")


class DatabaseConnectionError(ExpectedException):
    traceback = True
    halt = True

    def __init__(self, module_name=None):
        if module_name is not None:
            super().__init__(f"database connection error for {module_name}")
        else:
            super().__init__(f"database connection error")


class DatabaseError(ExpectedException):
    traceback = True
    halt = True

    def __init__(self, msg=None):
        if msg is not None:
            super().__init__(f"database error. {msg}")
        else:
            super().__init__(f"database error")


class ArgumentError(ExpectedException):
    halt = False
    traceback = True

    def __init__(self, msg=None):
        if msg is not None:
            super().__init__(f"{msg}")
        else:
            super().__init__(f"argument")


class WrongInput(ExpectedException):
    halt = False
    traceback = False

    def __init__(self, msg=None):
        if msg is not None:
            super().__init__(f"wrong input. {msg}")
        else:
            super().__init__(f"wrong input")


class NotAdmin(ExpectedException):
    halt = False
    traceback = False

    def __init__(self):
        super().__init__(f"No admin privilege")

# store-related exceptions
class ClientError(object):
    code = 0
    message = "Unspecified client error"


class InvalidModuleName(ClientError):
    code = 1
    message = "Invalid module name"


class InvalidVersionNumber(ClientError):
    code = 2
    message = "Invalid version number"


class WrongDeveloper(ClientError):
    code = 3
    message = "Developer does not have permission to edit this module"


class VersionExists(ClientError):
    code = 4
    message = "Version already exists"


class VersionDecrease(ClientError):
    code = 5
    message = "Version must increase"


class EmailUnverified(ClientError):
    code = 6
    message = "Email address unverified. Check your email for instructions to verify your email address"


class NoSuchModule(ClientError):
    code = 7
    message = "Module does not exist"


class AuthorizationError(ExpectedException):
    traceback = False
    halt = True

    def __init__(self):
        super().__init__(f"authorization failed")


class ServerError(Exception):
    def __init__(self):
        import traceback
        super().__init__(traceback.format_exc())
# end of store-related exceptions



from oakvar import BaseReporter
import sys


class Reporter(BaseReporter):
    def __init__(self, args):
        self.no_log = True
        self.no_status_update = True
        self.levels_to_write = None
        self.data = {}
        self.keep_json_all_mapping = True
        self.data = {}
        super().__init__(args)

    def write_preface(self, level):
        self.data[level] = []
        self.level = level

    def write_table_row(self, row):
        self.data[self.level].append(
            [row[col] for col in self.colnames_to_display[self.level]]
        )

    def end(self):
        info = {}
        norows = len(self.data[self.level])
        info["norows"] = norows
        self.data["info"] = info
        self.data["colinfo"] = self.colinfo
        self.data["warning_msgs"] = self.warning_msgs
        self.data["total_norows"] = self.total_norows
        return self.data


def main():
    reporter = Reporter(sys.argv)
    reporter.run()



from . import cli_entry
from . import cli_func


@cli_entry
@cli_func
def issue(__args__, __name__="issue"):
    from ..util.admin_util import report_issue

    return report_issue()


def get_parser_fn_issue():
    from argparse import ArgumentParser

    # opens issue report
    parser_fn_feedback = (
        ArgumentParser()
    )  # "feedback", help="opens a browser window to report issues")
    parser_fn_feedback.add_argument(
        "--quiet", action="store_true", default=None, help="run quietly"
    )
    parser_fn_feedback.set_defaults(func=issue)
    return parser_fn_feedback



from typing import Optional

job_statuses = {}


def unpack_eud(eud={}):
    email = eud.get("username")
    uid = eud.get("uid")
    dbpath = eud.get("dbpath")
    return email, uid, dbpath


def get_user_jobs_dir(request, email=None):
    from ...system import get_user_jobs_dir
    from .multiuser import get_email_from_request

    if not email:
        email = get_email_from_request(request)
    if not email:
        return None
    return get_user_jobs_dir(email)


async def get_job_dir_from_eud(_, eud={}) -> Optional[str]:
    from pathlib import Path
    from .serveradmindb import ServerAdminDb

    if eud.get("dbpath"):
        job_dir = Path(eud.get("dbpath")).parent
        if job_dir.exists():
            return str(job_dir)
        else:
            return None
    if eud.get("username") and eud.get("uid"):
        serveradmindb = ServerAdminDb()
        job_dir = await serveradmindb.get_job_dir_by_username_uid(eud=eud)
        return job_dir
    return None


async def get_user_job_run_name(_, eud={}) -> Optional[str]:
    from pathlib import Path
    from .multiuser import get_serveradmindb

    username, uid, dbpath = unpack_eud(eud=eud)
    if dbpath:
        return Path(dbpath).stem
    if username and uid:
        serveradmindb = await get_serveradmindb()
        run_name = await serveradmindb.get_run_name(username=username, uid=uid)
        return run_name
    return None


async def get_user_job_run_path(request, eud={}) -> Optional[str]:
    from pathlib import Path

    job_dir = await get_job_dir_from_eud(request, eud=eud)
    if job_dir is None:
        return None
    run_name = await get_user_job_run_name(request, eud=eud)
    if not run_name:
        return None
    run_path = Path(job_dir) / run_name
    return str(run_path)


async def get_user_job_report_paths(
    request, report_type: str, eud={}
) -> Optional[list]:
    from pathlib import Path
    from ...module.local import get_local_module_info_by_name

    run_path = await get_user_job_run_path(request, eud=eud)
    if not run_path:
        return None
    run_name = Path(run_path).name
    reporter = get_local_module_info_by_name(report_type + "reporter")
    if not reporter:
        return None
    output_filename_schema = reporter.conf.get("output_filename_schema")
    if not output_filename_schema:
        return None
    report_paths = []
    for pattern in output_filename_schema:
        report_paths.append(pattern.replace("{run_name}", run_name))
    return report_paths


async def get_user_job_dbpath(request, eud={}) -> Optional[str]:
    from ...consts import result_db_suffix

    if eud.get("dbpath"):
        return eud.get("dbpath")
    run_path = await get_user_job_run_path(request, eud=eud)
    if not run_path:
        return None
    dbpath = f"{run_path}{result_db_suffix}"
    return dbpath


async def get_user_job_log_path(request, eud={}) -> Optional[str]:
    from ...consts import LOG_SUFFIX

    run_path = await get_user_job_run_path(request, eud=eud)
    if not run_path:
        return None
    log_path = run_path + LOG_SUFFIX
    return log_path


def get_user_jobs_dir_list() -> Optional[list]:
    from pathlib import Path
    from ...system import get_jobs_dir

    user_jobs_dir_list = []
    root_jobs_dir = get_jobs_dir()
    for user_p in Path(root_jobs_dir).glob("*"):
        if not user_p.is_dir():
            continue
        user_jobs_dir_list.append(str(user_p.absolute()))
    return user_jobs_dir_list


def get_log_path_in_job_dir(
    job_dir: Optional[str], run_name: Optional[str] = None
) -> Optional[str]:
    from pathlib import Path
    from ...consts import LOG_SUFFIX

    if not job_dir:
        return None
    job_dir_p = Path(job_dir)
    if not job_dir_p.is_dir():
        return None
    if run_name:
        return str(job_dir_p / f"{run_name}{LOG_SUFFIX}")
    log_paths = [v for v in job_dir_p.glob("*.log")]
    return str(log_paths[0])


def get_job_runtime_in_job_dir(
    job_dir: Optional[str], run_name: Optional[str]
) -> Optional[float]:
    from pathlib import Path

    log_path = get_log_path_in_job_dir(job_dir, run_name=run_name)
    if not log_path:
        return None
    runtime = None
    if not Path(log_path).exists():
        return None
    with open(log_path) as f:
        for line in f:
            if "runtime:" in line:
                runtime = int(float(line.split("runtime:")[1].strip().rstrip("s")))
    return runtime



def list_modules(args):
    from oyaml import dump
    from ...module.remote import search_remote
    from ...module.local import search_local
    from ...module.local import get_local_module_info
    from ...module.remote import get_remote_module_info_ls
    from ...util.util import humanize_bytes
    from ...module.remote import RemoteModuleLs

    fmt = args.get("fmt", "return")
    nameonly = args.get("nameonly", False)
    available = args.get("available", False)
    types = args.get("types")
    tags = args.get("tags")
    all_toks = []
    all_toks_json = []
    if fmt == "tabular":
        if nameonly:
            all_toks = []
        else:
            if available:
                header = [
                    "Name",
                    "Title",
                    "Type",
                    "Size",
                    "Store version",
                    "Store data source",
                    "Installed",
                    "Local version",
                    "Local data source",
                ]
            else:
                header = [
                    "Name",
                    "Title",
                    "Type",
                    "Size",
                    "Version",
                    "Data source",
                ]
            all_toks = [header]
    elif fmt in ["json", "yaml"]:
        all_toks_json = []
    if available:
        if types:
            l = []
            for mt in types:
                l.extend(search_remote(args.get("pattern"), module_type=mt))
        else:
            l = search_remote(args.get("pattern"))
    else:
        l = search_local(args.get("pattern"))
    if l:
        for module_name in l:
            if available:
                module_info = get_remote_module_info_ls(module_name)
                if module_info:
                    add_local_module_info_to_remote_module_info(module_info)
            else:
                module_info = get_local_module_info(module_name)
            if not module_info:
                continue
            if types and module_info.type not in types:
                continue
            if tags:
                if not module_info.tags:
                    continue
                if not set(tags).intersection(module_info.tags):
                    continue
            if isinstance(module_info, RemoteModuleLs):
                size = module_info.size
            else:
                size = module_info.get_size()
            if not args.get("raw_bytes"):
                size = humanize_bytes(size)
            toks = []
            if fmt == "tabular":
                toks = [module_name]
                if not nameonly:
                    if available:
                        toks.extend(
                            [
                                module_info.title,
                                module_info.type,
                                size,
                                module_info.latest_code_version,
                                module_info.latest_data_source,
                                module_info.installed,
                                module_info.local_code_version,
                                module_info.local_data_source,
                            ]
                        )
                    else:
                        toks.extend(
                            [
                                module_info.title,
                                module_info.type,
                                size,
                                module_info.latest_code_version,
                                module_info.latest_data_source,
                            ]
                        )
                all_toks.append(toks)
            elif fmt in ["json", "yaml"]:
                toks = {"name": module_name}
                if not nameonly:
                    if available:
                        toks.update(
                            {
                                "title": module_info.title,
                                "type": module_info.type,
                                "size": size,
                                "version": module_info.latest_code_version,
                                "data_source": module_info.latest_data_source,
                                "installed": module_info.installed,
                                "local_code_version": module_info.local_code_version,
                                "local_data_source": module_info.local_data_source,
                            }
                        )
                    else:
                        toks.update(
                            {
                                "title": module_info.title,
                                "type": module_info.type,
                                "size": size,
                                "version": module_info.latest_code_version,
                                "data_source": module_info.latest_data_source,
                            }
                        )
                all_toks_json.append(toks)
    if fmt == "tabular":
        return all_toks
    elif fmt == "json":
        return all_toks_json
    elif fmt == "yaml":
        if all_toks_json:
            return dump(all_toks_json, default_flow_style=False)
        else:
            return ""
    else:
        return None


def add_local_module_info_to_remote_module_info(remote_info):
    from ...module.local import get_local_module_info

    local_info = get_local_module_info(remote_info.name)
    if local_info:
        remote_info.installed = "yes"
        remote_info.local_code_version = local_info.latest_code_version
        remote_info.local_data_source = local_info.latest_data_source
    else:
        remote_info.installed = ""
        remote_info.local_code_version = ""
        remote_info.local_data_source = ""





from typing import Dict
from typing import Tuple
from typing import Optional

ov_system_output_columns: Optional[dict] = None

def get_ucsc_bins(start, stop=None):
    if stop is None:
        stop = start + 1

    def range_per_level(start, stop):
        BIN_OFFSETS = [512 + 64 + 8 + 1, 64 + 8 + 1, 8 + 1, 1, 0]
        SHIFT_FIRST = 17
        SHIFT_NEXT = 3

        start_bin = start
        stop_bin = max(start, stop - 1)

        start_bin >>= SHIFT_FIRST
        stop_bin >>= SHIFT_FIRST

        for offset in BIN_OFFSETS:
            yield offset + start_bin, offset + stop_bin
            start_bin >>= SHIFT_NEXT
            stop_bin >>= SHIFT_NEXT

    return [
        x
        for first, last in range_per_level(start, stop)
        for x in range(first, last + 1)
    ]


def load_class(path, class_name=None):
    """Load a class from the class's name and path. (dynamic importing)"""
    from os.path import dirname, basename
    from importlib.util import spec_from_file_location, module_from_spec
    from importlib import import_module
    import sys
    import inspect
    from ..exceptions import ModuleLoadingError

    path_dir = dirname(path)
    sys.path = [path_dir] + sys.path
    module = None
    module_class = None
    module_name = basename(path).split(".")[0]
    try:
        module = import_module(module_name)
    except Exception as _:
        try:
            if class_name:
                spec = spec_from_file_location(class_name, path)
                if spec is not None:
                    module = module_from_spec(spec)
                    loader = spec.loader
                    if loader is not None:
                        loader.exec_module(module)
        except Exception as _:
            raise ModuleLoadingError(module_name)
    if module:
        if class_name:
            if hasattr(module, class_name):
                module_class = getattr(module, class_name)
                if not inspect.isclass(module_class):
                    module_class = None
        else:
            for n in dir(module):
                if n in [
                    "Converter",
                    "Mapper",
                    "Annotator",
                    "PostAggregator",
                    "Reporter",
                    "CommonModule",
                    "CravatConverter",
                    "CravatMapper",
                    "CravatAnnotator",
                    "CravatPostAggregator",
                    "CravatReporter",
                    "CravatCommonModule",
                ]:
                    if hasattr(module, n):
                        module_class = getattr(module, n)
                        if not inspect.isclass(module_class):
                            module_class = None
    del sys.path[0]
    return module_class


def get_directory_size(start_path):
    """
    Recursively get directory filesize.
    """
    from os import walk
    from os.path import join, getsize
    from ..store.consts import pack_ignore_fnames

    total_size = 0
    for dirpath, _, filenames in walk(start_path):
        for fname in filenames:
            if fname in pack_ignore_fnames:
                continue
            fp = join(dirpath, fname)
            total_size += getsize(fp)
    return total_size


def get_argument_parser_defaults(parser):
    defaults = {
        action.dest: action.default
        for action in parser._actions
        if action.dest != "help"
    }
    return defaults


def detect_encoding(path):
    from chardet.universaldetector import UniversalDetector
    from gzip import open as gzipopen

    if " " not in path:
        path = path.strip('"')
    if path.endswith(".gz"):
        f = gzipopen(path)
    else:
        f = open(path, "rb")
    detector = UniversalDetector()
    count = 0
    encoding = None
    for line in f:
        detector.feed(line)
        count += 1
        if detector.done or count == 100000:
            encoding = detector.result["encoding"]
            break
    detector.close()
    f.close()
    if not encoding:
        encoding = "utf-8"
    if encoding == "ascii":
        return "utf-8"
    else:
        return encoding


def get_job_version(dbpath, platform_name):
    from packaging.version import Version
    import sqlite3

    db = sqlite3.connect(dbpath)
    c = db.cursor()
    sql = f'select colval from info where colkey="{platform_name}"'
    c.execute(sql)
    r = c.fetchone()
    db_version = None
    if r is not None:
        db_version = Version(r[0])
    return db_version


def is_compatible_version(dbpath):
    from .admin_util import get_max_version_supported_for_migration
    from packaging.version import Version
    from pkg_resources import get_distribution

    max_version_supported_for_migration = get_max_version_supported_for_migration()
    try:
        ov_version = Version(get_distribution("oakvar").version)
    except:
        ov_version = None
    try:
        oc_version = Version(get_distribution("open-cravat").version)
    except:
        oc_version = None
    job_version_ov = get_job_version(dbpath, "oakvar")
    job_version_oc = get_job_version(dbpath, "open-cravat")
    if job_version_ov is None:
        if job_version_oc is None:
            compatible = False
        else:
            if job_version_oc < max_version_supported_for_migration:
                compatible = False
            else:
                compatible = True
        return compatible, job_version_oc, oc_version
    else:
        if job_version_ov < max_version_supported_for_migration:
            compatible = False
        else:
            compatible = True
        return compatible, job_version_ov, ov_version


def is_url(s) -> bool:
    if s.startswith("http://") or s.startswith("https://"):
        return True
    else:
        return False


def get_current_time_str():
    from datetime import datetime

    t = datetime.now()
    return t.strftime("%Y:%m:%d %H:%M:%S")


def get_args_conf(args: dict) -> Dict:
    if args is None:
        return {}
    # fill with run_conf dict
    run_conf = args.get("run_conf")
    if run_conf and type(run_conf) is dict:
        for k, v in run_conf.items():
            if k not in args or not args[k]:
                args[k] = v
    # fill with conf
    conf_path = args.get("confpath")
    if conf_path:
        conf = load_yml_conf(conf_path).get("run", {})
        args["conf"] = conf
        if conf:
            for k, v in conf.items():
                if k not in args or not args[k]:
                    args[k] = v
    return args


def get_args_package(args: dict) -> Dict:
    if args is None:
        return {}
    package = args.get("package")
    if package:
        from ..module.local import get_local_module_info

        m_info = get_local_module_info(package)
        if m_info:
            package_conf = m_info.conf
            package_conf_run = package_conf.get("run")
            if package_conf_run:
                for k, v in package_conf_run.items():
                    if k not in args or not args[k]:
                        args[k] = v
    return args


def get_args(parser, inargs, inkwargs):
    from types import SimpleNamespace
    from argparse import Namespace

    # given args. Combines arguments in various formats.
    inarg_dict = {}
    if inargs is not None:
        for inarg in inargs:
            t = type(inarg)
            if t == list:  # ['-t', 'text']
                inarg_dict.update(**vars(parser.parse_args(inarg)))
            elif t == Namespace:  # already parsed by a parser.
                inarg_dict.update(**vars(inarg))
            elif t == SimpleNamespace:
                inarg_dict.update(**vars(inarg))
            elif t == dict:  # {'output_dir': '/rt'}
                inarg_dict.update(inarg)
    if inkwargs is not None:
        inarg_dict.update(inkwargs)
    # conf
    inarg_dict = get_args_conf(inarg_dict)
    # package
    inarg_dict = get_args_package(inarg_dict)
    # defaults
    default_args = get_argument_parser_defaults(parser)
    for k, v in default_args.items():
        if k not in inarg_dict or inarg_dict[k] is None:
            inarg_dict[k] = v
    # convert value to list if needed.
    for action in parser._actions:
        if action.dest == "help":
            continue
        if action.nargs in ["+", "*"]:
            key = action.dest
            value = inarg_dict[key]
            if value and type(value) is not list:
                inarg_dict[key] = [value]
    return inarg_dict

def filter_affected_cols(filter):
    cols = set()
    if "column" in filter:
        cols.add(filter["column"])
    else:
        for rule in filter["rules"]:
            cols.update(filter_affected_cols(rule))
    return cols


def humanize_bytes(num, binary=False):
    """Human friendly file size"""
    from math import floor, log

    exp2unit_dec = {0: "B", 1: "kB", 2: "MB", 3: "GB"}
    exp2unit_bin = {0: "B", 1: "KiB", 2: "MiB", 3: "GiB"}
    max_exponent = 3
    if binary:
        base = 1024
    else:
        base = 1000
    if num > 0:
        exponent = floor(log(num, base))
        if exponent > max_exponent:
            exponent = max_exponent
    else:
        exponent = 0
    quotient = float(num) / base**exponent
    if binary:
        unit = exp2unit_bin[exponent]
    else:
        unit = exp2unit_dec[exponent]
    quot_str = "{:.1f}".format(quotient)
    # No decimal for byte level sizes
    if exponent == 0:
        quot_str = quot_str.rstrip("0").rstrip(".")
    return "{quotient} {unit}".format(quotient=quot_str, unit=unit)


def get_dict_from_namespace(n):
    from types import SimpleNamespace
    from argparse import Namespace

    if type(n) == SimpleNamespace or type(n) == Namespace:
        n = vars(n)
    return n


def quiet_print(msg, args=None, quiet=None):
    from .util import get_dict_from_namespace

    args = get_dict_from_namespace(args)
    if quiet is None:
        if args and args.get("quiet") is not None:
            quiet = args.get("quiet")
        else:
            quiet = True
    if quiet == False:
        if args:
            outfn = args.get("outfn", print)
        else:
            outfn = print
        outfn(msg, flush=True)


def email_is_valid(email: str) -> bool:
    from re import fullmatch

    if fullmatch(
        r"([A-Za-z0-9]+[.-_])*[A-Za-z0-9]+@[A-Za-z0-9-]+(\.[A-Z|a-z]{2,})+", email
    ):
        return True
    else:
        return False


def pw_is_valid(pw: str) -> bool:
    from re import fullmatch

    if fullmatch(r"[a-zA-Z0-9!?@*&\-\+]+", pw):
        return True
    else:
        return False


def load_yml_conf(yml_conf_path):
    from oyaml import safe_load

    with open(yml_conf_path, encoding="utf-8") as f:
        conf = safe_load(f)
    if conf == None:
        conf = {}
    return conf


def compare_version(v1, v2):
    from packaging.version import Version

    sv1 = Version(v1)
    sv2 = Version(v2)
    if sv1 == sv2:
        return 0
    elif sv1 > sv2:
        return 1
    else:
        return -1


def get_email_pw_from_input(email=None, pw=None, pwconfirm=False) -> Tuple[str, str]:
    from getpass import getpass

    email = email or ""
    pw = pw or ""
    done = False
    startover = False
    while not done:
        while not email or startover:
            startover = False
            if email:
                email = input(f"e-mail [{email}]: ") or email
            else:
                email = input(f"e-mail: ")
            if not email_is_valid(email):
                print(f"invalid e-mail")
                email = None
        while not pw and not startover:
            pw = getpass("Password (or just enter to start over): ")
            if pw == "":
                startover = True
                break
            if not pw_is_valid(pw):
                print(f"invalid password")
                pw = None
        if pwconfirm:
            pwagain = None
            while not pwagain and not startover:
                pwagain = getpass("Password again (or just enter to start over): ")
                if pwagain == "":
                    pw = ""
                    startover = True
                    break
                if not pw_is_valid(pwagain):
                    print(f"invalid password")
                    pwagain = None
                elif pw != pwagain:
                    print(f"password mismatch")
                    pwagain = None
                elif pw == pwagain:
                    done = True
                    break
        else:
            done = True
    print("")
    return email, pw


def get_email_from_args(args={}) -> Optional[str]:
    return args.get("email")


def version_requirement_met(version, target_version) -> bool:
    from packaging.version import Version

    if not target_version:
        return True
    return Version(version) >= Version(target_version)


def get_latest_version(versions, target_version=None):
    from packaging.version import Version

    latest_version = ""
    for version in versions:
        if not version_requirement_met(version, target_version):
            continue
        if not latest_version or Version(version) > Version(latest_version):
            latest_version = version
    return latest_version


def escape_glob_pattern(pattern):
    new_pattern = "[[]".join(["[]]".join(v.split("]")) for v in pattern.split("[")])
    return new_pattern.replace("*", "[*]").replace("?", "[?]")


def get_random_string(k=16):
    from random import choices
    from string import ascii_lowercase

    return "".join(choices(ascii_lowercase, k=k))


def get_result_dbpath(output_dir: str, run_name: str):
    from pathlib import Path
    from ..consts import result_db_suffix

    return str(Path(output_dir) / (run_name + result_db_suffix))

def get_unique_path(path: str):
    from pathlib import Path

    count = 1
    p = Path(path)
    stem = p.stem
    suffix = p.suffix
    while p.exists():
        p = Path(f"{stem}_{count}{suffix}")
    return str(p)


def yield_tabular_lines(l, col_spacing=2, indent=0):
    if not l:
        return
    sl = []
    n_toks = len(l[0])
    max_lens = [0] * n_toks
    for toks in l:
        if len(toks) != n_toks:
            raise RuntimeError("Inconsistent sub-list length")
        stoks = [str(x) for x in toks]
        sl.append(stoks)
        stoks_len = [len(x) for x in stoks]
        max_lens = [max(x) for x in zip(stoks_len, max_lens)]
    for stoks in sl:
        jline = " " * indent
        for i, stok in enumerate(stoks):
            jline += stok + " " * (max_lens[i] + col_spacing - len(stok))
        yield jline


def print_tabular_lines(l, args=None):
    for line in yield_tabular_lines(l):
        if args:
            quiet_print(line, args=args)
        else:
            print(line)

def log_module(module, logger):
    if logger:
        code_version = None
        if hasattr(module, "conf"):
            code_version = module.conf.get("code_version") or module.conf.get("version")
        if not code_version and hasattr(module, "version"):
            code_version = module.version
        if not code_version:
            code_version = "?"
        logger.info(f"module: {module.name}=={code_version} {module.script_path}")

def get_ov_system_output_columns() -> dict:
    from pathlib import Path
    from oyaml import safe_load

    global ov_system_output_columns
    if ov_system_output_columns:
        return ov_system_output_columns
    fp = Path(__file__).parent.parent / "assets" / "output_columns.yml"
    with open(fp) as f:
        ov_system_output_columns = safe_load(f)
    if not ov_system_output_columns:
        return {}
    return ov_system_output_columns

def get_crv_def() -> list:
    from ..consts import INPUT_LEVEL_KEY

    output_columns: dict = get_ov_system_output_columns()
    return output_columns[INPUT_LEVEL_KEY]

def get_crx_def() -> list:
    from ..consts import VARIANT_LEVEL_KEY

    output_columns: dict = get_ov_system_output_columns()
    return output_columns[VARIANT_LEVEL_KEY]

def get_crg_def() -> list:
    from ..consts import GENE_LEVEL_KEY

    output_columns: dict = get_ov_system_output_columns()
    return output_columns[GENE_LEVEL_KEY]

def get_crs_def() -> list:
    from ..consts import SAMPLE_LEVEL_KEY

    output_columns: dict = get_ov_system_output_columns()
    return output_columns[SAMPLE_LEVEL_KEY]

def get_crm_def() -> list:
    from ..consts import MAPPING_LEVEL_KEY

    output_columns: dict = get_ov_system_output_columns()
    return output_columns[MAPPING_LEVEL_KEY]

def get_crl_def() -> list:
    from ..consts import LIFTOVER_LEVEL_KEY

    output_columns: dict = get_ov_system_output_columns()
    return output_columns[LIFTOVER_LEVEL_KEY]




from typing import Optional
from typing import Any
from aiohttp import web
from aiohttp import web_runner

class WebServer(object):
    def __init__(self, loop=None, url=None, args={}):
        from ..util.asyn import get_event_loop

        self.args = args
        self.logger = args.get("logger")
        self.app = None
        self.cors = Optional[Any]
        self.runner = None
        self.site = None
        self.host = args.get("host")
        self.port = args.get("port")
        if loop is None:
            loop = get_event_loop()
        self.ssl_context = args.get("ssl_context")
        self.loop = loop
        self.server_started = False
        task = loop.create_task(self.start())
        task.add_done_callback(self.server_done)
        if args.get("headless") == False and url:
            self.loop.create_task(self.open_url(url))

    @web.middleware
    async def middleware(self, request, handler):
        from json import dumps
        from traceback import print_exc
        from sys import stderr
        from aiohttp.web_exceptions import HTTPNotFound

        global loop
        try:
            url_parts = request.url.parts
            response = await handler(request)
            nocache = False
            if url_parts[0] == "/":
                if len(url_parts) >= 3 and url_parts[2] == "nocache":
                    nocache = True
            elif url_parts[0] == "nocache":
                nocache = True
            if nocache:
                response.headers["Cache-Control"] = "no-cache"
                response.headers["Access-Control-Allow-Origin"] = "*"
            return response
        except Exception as e:
            msg = "Exception with {}".format(request.rel_url)
            if self.logger:
                self.logger.info(msg)
                self.logger.exception(e)
            if self.args.get("debug"):
                if not isinstance(e, HTTPNotFound):
                    stderr.write(msg + "\n")
                    print_exc()
            return web.HTTPInternalServerError(
                text=dumps({"status": "error", "msg": str(e)})
            )

    def server_done(self, task):
        from traceback import print_exc

        try:
            task.result()
        except Exception as e:
            if self.logger:
                self.logger.exception(e)
            if self.args.get("debug"):
                print_exc()
            self.server_started = None
            exit(1)

    async def open_url(self, url):
        from webbrowser import open as webbrowseropen
        from asyncio import sleep

        while self.server_started == False:
            await sleep(0.2)
        if self.server_started:
            webbrowseropen(url)

    def start_workers(self):
        from .webstore import webstore as ws
        from .websubmit import websubmit as wu

        ws.start_worker()
        wu.start_worker()

    async def start(self):
        global server_ready
        from aiohttp import web
        import aiohttp_cors

        self.app = web.Application(loop=self.loop, middlewares=[self.middleware], client_max_size=1024*1024*1024*1024)
        self.cors = aiohttp_cors.setup(self.app) # type: ignore
        self.setup_routes()
        self.start_workers()
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = TCPSitePatched(
            self.runner,
            host=self.host,
            port=self.port,
            loop=self.loop,
            ssl_context=self.ssl_context,
        )
        await self.site.start()
        self.server_started = True

    def setup_webapp_routes(self):
        from ..exceptions import ModuleLoadingError
        from importlib.util import spec_from_file_location, module_from_spec
        from os.path import join, exists
        from os import listdir
        from ..exceptions import ModuleLoadingError
        from ..exceptions import SetupError
        from ..system import get_modules_dir

        modules_dir = get_modules_dir()
        if modules_dir is None:
            return False
        webapps_dir = join(modules_dir, "webapps")
        if exists(webapps_dir) == False:
            return False
        module_names = listdir(webapps_dir)
        for module_name in module_names:
            try:
                module_dir = join(webapps_dir, module_name)
                pypath = join(module_dir, "route.py")
                if exists(pypath):
                    spec = spec_from_file_location("route", pypath)
                    if spec is None:
                        raise ModuleLoadingError(module_name)
                    module = module_from_spec(spec)
                    if spec.loader is None:
                        raise ModuleLoadingError(module_name)
                    spec.loader.exec_module(module)
                    if self.app is None:
                        raise SetupError()
                    for route in module.routes:
                        method, path, func_name = route
                        path = f"/webapps/{module_name}/" + path
                        self.app.router.add_route(method, path, func_name)
            except Exception as e:
                if self.logger:
                    self.logger.error(f"error loading webapp {module_name}")
                    self.logger.exception(e)
        return True

    def setup_routes(self):
        from .webresult import webresult as wr
        from .webstore import webstore as ws
        from .websubmit import websubmit as wu
        from .websubmit import multiuser as mu
        from ..exceptions import SetupError
        from ..system import get_modules_dir
        from os.path import dirname
        from os.path import realpath
        from os.path import join
        from os.path import exists
        from aiohttp_cors import ResourceOptions

        if self.app is None:
            raise SetupError()
        source_dir = dirname(realpath(__file__))
        routes = list()
        routes.extend(ws.routes)
        routes.extend(wr.routes)
        routes.extend(wu.routes)
        mu.add_routes(self.app.router)
        for route in routes:
            method, path, func_name = route
            self.app.router.add_route(method, path, func_name)
        self.setup_webapp_routes()
        self.app.router.add_static("/store", join(source_dir, "..", "gui", "webstore"))
        self.app.router.add_static(
            "/result", join(source_dir, "..", "gui", "webresult")
        )
        self.app.router.add_static("/submit", join(source_dir, "websubmit"))
        self.app.router.add_static("/", join(source_dir, "www"))
        modules_dir = get_modules_dir()
        if modules_dir:
            if exists(join(modules_dir, "annotators")):
                self.app.router.add_static(
                    "/modules/annotators/", join(modules_dir, "annotators")
                )
            if exists(join(modules_dir, "webapps")):
                self.app.router.add_static("/webapps", join(modules_dir, "webapps"))
        if self.cors:
            for resource in list(self.app.router.resources()):
                self.cors.add(
                    resource,
                    {
                        "*": ResourceOptions(
                            allow_credentials=True,
                            expose_headers="*",
                            allow_headers="*",
                            allow_methods="*",
                        ),
                    },
                )


class TCPSitePatched(web_runner.BaseSite):
    __slots__ = ("loop", "_host", "_port", "_reuse_address", "_reuse_port")

    def __init__(
        self,
        runner,
        host=None,
        port=None,
        shutdown_timeout=60.0,
        ssl_context=None,
        backlog=128,
        reuse_address=None,
        reuse_port=None,
        loop=None,
    ):
        from ..util.asyn import get_event_loop

        super().__init__(
            runner,
            shutdown_timeout=shutdown_timeout,
            ssl_context=ssl_context,
            backlog=backlog,
        )
        if loop is None:
            loop = get_event_loop()
        self.loop = loop
        if host is None:
            host = "0.0.0.0"
        self._host = host
        if port is None:
            port = 8443 if self._ssl_context else 8060
        self._port = port
        self._reuse_address = reuse_address
        self._reuse_port = reuse_port

    @property
    def name(self, args={}):
        from yarl import URL

        scheme = "https" if args.get("ssl_enabled", False) else "http"
        return str(URL.build(scheme=scheme, host=self._host, port=self._port))

    async def start(self):
        from ..exceptions import SetupError

        await super().start()
        if self._runner.server is None:
            raise SetupError()
        self._server = await self.loop.create_server(
            self._runner.server,
            self._host,
            self._port,
            ssl=self._ssl_context,  # type: ignore
            backlog=self._backlog,
            reuse_address=self._reuse_address,
            reuse_port=self._reuse_port,
        )



def update_status(status: str, logger=None, serveradmindb=None):
    if logger:
        logger.info(status)
    if serveradmindb:
        serveradmindb.update_job_info({"status": status})


def announce_module(module, logger=None, serveradmindb=None):
    update_status(
        f"running {module.name}",
        logger=logger,
        serveradmindb=serveradmindb,
    )


def log_variant_exception(
    lnum=0,
    line="",
    __input_data__=None,
    unique_excs=[],
    logger=None,
    error_logger=None,
    e=None,
):
    import traceback

    if logger:
        err_str = traceback.format_exc().rstrip()
        if err_str.endswith("None"):
            err_str_log = str(e)
        else:
            err_str_log = err_str
        if not err_str_log in unique_excs:
            unique_excs.append(err_str_log)
            logger.error(err_str_log)
    if error_logger:
        error_logger.error("\n[{:d}]{}\n({})\n#".format(lnum, line.rstrip(), str(e)))


def print_log_handlers():
    import logging

    for k, v in logging.Logger.manager.loggerDict.items():
        if "oakvar" in k:
            print("+ [%s] {%s} " % (str.ljust(k, 20), str(v.__class__)[8:-2]))
            if not isinstance(v, logging.PlaceHolder):
                for h in v.handlers:
                    print("     +++", str(h.__class__)[8:-2])
                    for fld, val in h.__dict__.items():
                        print("%s%s=%s" % ("   -", fld, val))


def get_y_or_n():
    while True:
        resp = input("Proceed? ([y]/n) > ")
        if resp == "y" or resp == "":
            return True
        if resp == "n":
            return False
        else:
            continue


def show_logo():
    print(
        r"""
==========================================================
 #######     ###    ##    ## ##     ##    ###    ########  
##     ##   ## ##   ##   ##  ##     ##   ## ##   ##     ## 
##     ##  ##   ##  ##  ##   ##     ##  ##   ##  ##     ## 
##     ## ##     ## #####    ##     ## ##     ## ########  
##     ## ######### ##  ##    ##   ##  ######### ##   ##   
##     ## ##     ## ##   ##    ## ##   ##     ## ##    ##  
 #######  ##     ## ##    ##    ###    ##     ## ##     ##
==========================================================
                                   Oak Bioinformatics, LLC
              Licensed under AGPL-3 and commercial license
        Licensing and feedback: info@oakbioinformatics.com
                                        https://oakvar.com
""",
        flush=True,
    )

def get_module_options(args):
    from json import loads
    from ..consts import MODULE_OPTIONS_KEY

    module_options = args.get(MODULE_OPTIONS_KEY)
    if module_options:
        if isinstance(module_options, str):
            module_options = args.get(MODULE_OPTIONS_KEY).lstrip("'").rstrip("'").replace("'", '"')
            module_options = loads(module_options)
        elif not isinstance(module_options, dict):
            module_options = None
    return module_options


def get_new_job_dir(jobs_dir: str) -> str:
    from datetime import datetime
    from pathlib import Path

    job_name = datetime.now().strftime(r"%y%m%d-%H%M%S")
    job_dir = Path(jobs_dir) / job_name
    if job_dir.exists():
        count = 1
        while True:
            job_name = datetime.now().strftime(r"%y%m%d-%H%M%S") + "_" + str(count)
            job_dir = Path(jobs_dir) / job_name
            if not job_dir.exists():
                break
            count += 1
    return str(job_dir)


def get_new_job_name(jobs_dir: str) -> str:
    from pathlib import Path

    job_dir = get_new_job_dir(jobs_dir)
    job_name = Path(job_dir).name
    return job_name



def print_module_info(module_info={}):
    from rich.console import Console

    console = Console()
    readme_table = get_module_info_readme_table(module_info=module_info)
    basic_table = get_module_info_basic_table(module_info=module_info)
    developer_table = get_module_info_developer_table(module_info=module_info)
    version_table = get_module_info_version_table(module_info=module_info)
    output_table = get_module_info_output_table(module_info=module_info)
    console.print(readme_table)
    console.print(basic_table)
    console.print(developer_table)
    console.print(version_table)
    console.print(output_table)


def get_module_info_readme_table(module_info={}):
    from rich.table import Table
    from rich import box

    readme = module_info.get("readme")
    readme_table = Table(
        title="README", title_style="bold", show_header=False, box=box.SQUARE
    )
    readme_table.add_column("Readme")
    readme_table.add_row(readme)
    return readme_table


def get_module_info_basic_table(module_info={}):
    from rich.table import Table
    from rich import box

    table = Table(show_header=False, title_style="bold", box=box.SQUARE)
    table.add_column("Category")
    table.add_column("Value")
    for k, v in module_info.items():
        if k in ["readme", "output_columns", "versions", "developer"]:
            continue
        if type(v) == list:
            v = ", ".join(v)
        table.add_row(k, str(v))
    return table


def add_module_info_developer_table_rows(developer_table, developers):
    developer_table.add_row("Name", str(developers.get("name")))
    developer_table.add_row("Organization", str(developers.get("organization")))
    developer_table.add_row("Email", str(developers.get("email")))
    developer_table.add_row("Website", str(developers.get("website")))
    developer_table.add_row("Citation", str(developers.get("citation")))


def get_module_info_developer_table(module_info={}):
    from rich.table import Table
    from rich import box

    developers = module_info.get("developer")
    developer_table = Table(
        title="Developers", title_style="bold", show_header=False, box=box.SQUARE
    )
    developer_table.add_column("Category")
    developer_table.add_column("Value")
    if "name" in developers:
        add_module_info_developer_table_rows(developer_table, developers)
    else:
        if "module" in developers:
            developer_table.add_row("[bold]Module[/bold]", "")
            add_module_info_developer_table_rows(
                developer_table, developers.get("module")
            )
        if "data" in developers:
            developer_table.add_row("[bold]Data[/bold]", "")
            add_module_info_developer_table_rows(
                developer_table, developers.get("data")
            )
    return developer_table


def get_module_info_version_table(module_info={}):
    from rich.table import Table
    from rich import box
    from packaging.version import Version

    versions = module_info.get("versions")
    version_table = Table(title="Versions", title_style="bold", box=box.SQUARE)
    version_table.add_column("Version")
    version_table.add_column("Data version")
    version_table.add_column("Data source")
    version_table.add_column("Minimum OakVar version")
    code_vers = [v for v in versions.keys()]
    code_vers.sort(key=lambda x: Version(x))
    for code_ver in code_vers:
        dd = versions.get(code_ver)
        version_table.add_row(code_ver, dd.get("data_version"), dd.get("data_source"), dd.get("min_pkg_ver"))
    return version_table


def get_module_info_output_table(module_info={}):
    from rich.table import Table
    from rich import box

    output_columns = module_info.get("output_columns")
    output_table = Table(title="Output", title_style="bold", box=box.SQUARE)
    output_table.add_column("Name")
    output_table.add_column("Title")
    output_table.add_column("Description")
    output_table.add_column("Type")
    for col in output_columns:
        ty = col.get("type", "string")
        output_table.add_row(col.get("name"), col.get("title"), col.get("desc"), ty)
    return output_table






from typing import Optional

pkg_version: Optional[str] = None

class ReadyState(object):

    READY = 0
    MISSING_MD = 1
    UPDATE_NEEDED = 2
    NO_BASE_MODULES = 3

    messages = {
        0: "",
        1: "Modules directory not found",
        2: 'Update on system modules needed. Run "oc module install-base"',
        3: "Base modules do not exist.",
    }

    def __init__(self, code=READY):
        if code not in self.messages:
            raise ValueError(code)
        self.code = code

    @property
    def message(self):
        return self.messages[self.code]

    def __bool__(self):
        return self.code == self.READY

    def __iter__(self):
        yield "ready", bool(self)
        yield "code", self.code
        yield "message", self.message


def get_user_conf():
    from ..system import get_user_conf_path
    from ..util.util import load_yml_conf
    from os.path import exists

    conf_path = get_user_conf_path()
    if exists(conf_path):
        ret = load_yml_conf(conf_path)
        ret["conf_path"] = conf_path
        return ret
    else:
        return None


def get_current_package_version():
    from pkg_resources import get_distribution

    version = get_distribution("oakvar").version
    return version


def get_default_assembly():
    conf = get_user_conf()
    if conf:
        default_assembly = conf.get("default_assembly", None)
        return default_assembly


def get_last_assembly():
    conf = get_user_conf()
    if conf:
        last_assembly = conf.get("last_assembly")
        return last_assembly


def get_latest_package_version():
    """
    Return latest oakvar version on pypi
    """
    all_vers = get_package_versions()
    if all_vers:
        return all_vers[-1]
    else:
        return None


def get_package_versions():
    """
    Return available oakvar versions from pypi, sorted asc
    """
    import json
    from requests import get
    from requests.exceptions import ConnectionError
    from packaging.version import Version
    from ..exceptions import InternetConnectionError

    try:
        r = get("https://pypi.org/pypi/oakvar/json", timeout=(3, None))
    except ConnectionError:
        raise InternetConnectionError()
    if r.status_code == 200:
        d = json.loads(r.text)
        all_vers = list(d["releases"].keys())
        all_vers.sort(key=Version)
        return all_vers
    else:
        return None


async def get_updatable_async(modules=[], strategy="consensus"):
    from ..module import get_updatable

    update_vers, resolution_applied, resolution_failed = get_updatable(
        modules=modules, strategy=strategy
    )
    return [update_vers, resolution_applied, resolution_failed]


def get_widgets_for_annotator(annotator_name, skip_installed=False):
    from ..module.local import module_exists_local
    from ..module.remote import get_remote_module_info
    from ..module import list_remote
    from ..module.cache import get_module_cache

    linked_widgets = []
    l = list_remote()
    if not l:
        return None
    for widget_name in l:
        widget_info = get_remote_module_info(widget_name)
        if widget_info is not None and widget_info.type == "webviewerwidget":
            widget_config = get_module_cache().get_remote_module_piece_url(
                widget_name, "config"
            )
            if widget_config:
                linked_annotator = widget_config.get("required_annotator")
            else:
                linked_annotator = None
            if linked_annotator == annotator_name:
                if skip_installed and module_exists_local(widget_name):
                    continue
                else:
                    linked_widgets.append(widget_info)
    return linked_widgets


def input_formats():
    import os
    from ..system import get_modules_dir

    formats = set()
    d = os.path.join(get_modules_dir(), "converters")
    if os.path.exists(d):
        fns = os.listdir(d)
        for fn in fns:
            if fn.endswith("-converter"):
                formats.add(fn.split("-")[0])
    return formats


def install_widgets_for_module(module_name):
    from ..module import install_module

    widget_name = "wg" + module_name
    install_module(widget_name)


def fn_new_exampleinput(d):
    import shutil
    import os

    fn = "exampleinput"
    ifn = os.path.join(get_packagedir(), fn)
    ofn = os.path.join(d, fn)
    shutil.copyfile(ifn, ofn)
    return ofn


def new_annotator(annot_name):
    import shutil
    import os
    from ..system import get_modules_dir
    from ..module.cache import get_module_cache

    annot_root = os.path.join(get_modules_dir(), "annotators", annot_name)
    template_root = os.path.join(get_packagedir(), "annotator_template")
    shutil.copytree(template_root, annot_root)
    for dir_path, _, fnames in os.walk(annot_root):
        for old_fname in fnames:
            old_fpath = os.path.join(dir_path, old_fname)
            new_fname = old_fname.replace("annotator_template", annot_name, 1)
            new_fpath = os.path.join(dir_path, new_fname)
            os.rename(old_fpath, new_fpath)
    get_module_cache().update_local()


def ready_resolution_console():
    return system_ready()


def recursive_update(d1, d2):
    """
    Recursively merge two dictionaries and return a copy.
    d1 is merged into d2. Keys in d1 that are not present in d2 are preserved
    at all levels. The default Dict.update() only preserved keys at the top
    level.
    """
    import copy

    d3 = copy.deepcopy(d1)  # Copy perhaps not needed. Test.
    for k, v in d2.items():
        if k in d3:
            orig_v = d3[k]
            if isinstance(v, dict):
                if isinstance(orig_v, dict) == False:
                    d3[k] = v
                else:
                    t = recursive_update(d3.get(k, {}), v)
                    d3[k] = t
            else:
                d3[k] = d2[k]
        else:
            d3[k] = v
    return d3


def report_issue():
    import webbrowser

    webbrowser.open("http://github.com/rkimoakbioinformatics/oakvar/issues")


def set_user_conf_prop(key, val):
    import oyaml as yaml
    from ..system import get_user_conf_path

    conf = get_user_conf()
    if conf:
        conf[key] = val
        wf = open(get_user_conf_path(), "w")
        yaml.dump(conf, wf, default_flow_style=False)
        wf.close()


def set_jobs_dir(d):
    from ..system import update_system_conf_file

    update_system_conf_file({"jobs_dir": d})


# return a list of module types (e.g. annotators) in the local install
def show_main_conf(args):
    import oyaml as yaml
    from ..util.util import quiet_print

    conf = get_user_conf()
    if args["fmt"] == "yaml":
        conf = yaml.dump(conf, default_flow_style=False)
    if args["to"] == "stdout":
        quiet_print(conf, args=args)
    else:
        return conf


def oakvar_version():
    global pkg_version
    if not pkg_version:
        pkg_version = get_current_package_version()
    return pkg_version


def system_ready():
    import os
    from ..system import get_modules_dir
    from ..exceptions import NoModulesDir
    from ..exceptions import NoSystemModule

    modules_dir = get_modules_dir()
    if not (os.path.exists(modules_dir)):
        raise NoModulesDir()
    elif not (os.path.exists(os.path.join(modules_dir, "converters", "vcf-converter"))):
        raise NoSystemModule()
    else:
        return ReadyState()


def write_user_conf(user_conf):
    import oyaml as yaml
    from ..system import get_user_conf_path

    confpath = get_user_conf_path()
    wf = open(confpath, "w")
    yaml.dump(user_conf, wf, default_flow_style=False)
    wf.close()


def get_liftover_chain_paths():
    from os.path import join

    liftover_chains_dir = get_liftover_chains_dir()
    liftover_chain_paths = {
        "hg19": join(liftover_chains_dir, "hg19ToHg38.over.chain"),
        "hg18": join(liftover_chains_dir, "hg18ToHg38.over.chain"),
    }
    return liftover_chain_paths


def get_packagedir():
    import os.path

    return os.path.abspath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")
    )


def get_platform():
    from platform import platform

    pl = platform()
    if pl.startswith("Windows"):
        pl = "windows"
    elif pl.startswith("Darwin") or pl.startswith("macOS"):
        pl = "macos"
    elif pl.startswith("Linux"):
        pl = "linux"
    else:
        pl = "linux"
    return pl


def get_liftover_chains_dir():
    from os.path import join as pathjoin

    return pathjoin(get_packagedir(), "liftover")


def get_max_version_supported_for_migration():
    from packaging.version import Version

    return Version("1.7.0")



from . import cli_entry
from . import cli_func

protocol = None
log_path = None
loop = None

def inject_module_variables(args={}):
    from ..gui.webresult import webresult as wr
    from ..gui.webstore import webstore as ws
    from ..gui.websubmit import websubmit as wu
    from ..gui.websubmit import multiuser as mu

    logger = args.get("logger")
    wu.logger = logger
    ws.logger = logger
    wr.logger = logger
    mu.logger = logger
    wu.servermode = args.get("servermode")
    ws.servermode = args.get("servermode")
    wr.servermode = args.get("servermode")
    mu.servermode = args.get("servermode")
    wr.wu = wu
    wu.mu = mu
    ws.mu = mu


def get_protocol(args={}):
    global protocol
    if not protocol:
        if args.get("ssl_enabled"):
            protocol = "https://"
        else:
            protocol = "http://"
    return protocol


def setup(args={}):
    from os.path import abspath
    from ..gui.websubmit import multiuser as mu
    from ..util.asyn import get_event_loop

    if args.get("result"):
        args["headless"] = False
        args["result"] = abspath(args.get("result"))
    elif args.get("servermode"):
        args["headless"] = True
    inject_module_variables(args=args)
    loop = get_event_loop()
    loop.run_until_complete(mu.get_serveradmindb())
    args["ssl_enabled"] = False


def is_port_occupied(args={}):
    import socket
    from ..gui.util import get_host_port

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(1)
    host, port = get_host_port(args=args)
    sr = -1
    try:
        sr = s.connect_ex((host, port))
        s.close()
    except ConnectionError:
        pass
    return sr == 0


def get_pem_path(args={}):
    from os.path import join

    sysconf = args.get("sysconf", {})
    if "conf_dir" in sysconf:
        pem_path = join(sysconf.get("conf_dir"), "cert.pem")
    else:
        pem_path = None
    return pem_path


def get_ssl_context(args={}):
    from os.path import exists
    from ssl import create_default_context, Purpose

    pem_path = get_pem_path(args=args)
    if pem_path and exists(pem_path) and args.get("http_only") == False:
        sc = create_default_context(Purpose.CLIENT_AUTH)
        sc.load_cert_chain(pem_path)
    else:
        sc = None
    return sc


def main(url=None, args={}):
    from webbrowser import open as open_browser
    from ..util.util import quiet_print
    from ..util.run import show_logo
    from ..gui.server import WebServer
    from ..gui.util import get_host_port
    from ..util.asyn import get_event_loop

    logger = args.get("logger")
    if is_port_occupied(args=args):
        msg = (
            f"OakVar or another program is already running at port {args.get('port')}."
        )
        if logger:
            logger.info(msg)
        quiet_print(msg, args)
        if url and not args.get("headless"):
            open_browser(url)
        return
    show_logo()
    host, port = get_host_port(args=args)
    msg = f"OakVar Server is served at {host}:{port}"
    if logger:
        logger.info(msg)
    quiet_print(msg, args=args)
    quiet_print("(To quit: Press Ctrl-C or Ctrl-Break)", args=args)
    loop = get_event_loop()
    if args["ssl_enabled"]:
        args["ssl_context"] = get_ssl_context(args=args)
    _ = WebServer(loop=loop, url=url, args=args)
    loop.run_forever()


def get_parser_fn_gui():
    from argparse import ArgumentParser

    parser_fn_gui = ArgumentParser()
    parser_fn_gui.add_argument(
        "--multiuser",
        dest="servermode",
        action="store_true",
        default=False,
        help="Runs in multiuser mode",
    )
    parser_fn_gui.add_argument(
        "--headless",
        action="store_true",
        default=False,
        help="do not open the OakVar web page",
    )
    parser_fn_gui.add_argument(
        "--http-only",
        action="store_true",
        default=False,
        help="Force not to accept https connection",
    )
    parser_fn_gui.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Console echoes exceptions written to log file.",
    )
    parser_fn_gui.add_argument(
        "result", nargs="?", help="Path to a OakVar result SQLite file"
    )
    parser_fn_gui.add_argument(
        "--webapp",
        dest="webapp",
        default=None,
        help="Name of OakVar webapp module to run",
    )
    parser_fn_gui.add_argument(
        "--port",
        dest="port",
        default=None,
        help="Port number for OakVar graphical user interface",
    )
    parser_fn_gui.add_argument(
        "--noguest",
        dest="noguest",
        default=False,
        action="store_true",
        help="Disables guest mode",
    )
    parser_fn_gui.add_argument(
        "--quiet", action="store_true", default=None, help="run quietly"
    )
    parser_fn_gui.add_argument("--to", default="return", help="run quietly")
    parser_fn_gui.set_defaults(func=cli_gui)
    return parser_fn_gui


def get_logger(args={}):
    from logging import getLogger
    from logging import INFO
    from logging import Formatter
    from logging.handlers import TimedRotatingFileHandler
    from ..system.consts import log_dir_key
    from ..gui.util import get_log_path

    global log_path
    log_dir = args.get("sysconf", {}).get(log_dir_key)
    log_path = get_log_path(log_dir=log_dir)
    logger = getLogger()
    logger.setLevel(INFO)
    log_handler = TimedRotatingFileHandler(log_path, when="d", backupCount=30)
    log_formatter = Formatter("%(asctime)s: %(message)s", "%Y/%m/%d %H:%M:%S")
    log_handler.setFormatter(log_formatter)
    logger.addHandler(log_handler)
    return logger


def get_webapp_url(args={}):
    from os.path import join
    from os.path import exists
    from sys import stderr
    from ..system.consts import modules_dir_key
    from ..gui.util import get_host_port

    host, port = get_host_port(args=args)
    sysconf = args.get("sysconf", {})
    index_path = join(
        sysconf.get(modules_dir_key), "webapps", args["webapp"], "index.html"
    )
    if exists(index_path) == False:
        stderr.write(f"Webapp {args['webapp']} does not exist. Exiting.\n")
        return
    url = f"{host}:{port}/webapps/{args['webapp']}/index.html"
    return url


def get_result_url(args={}):
    from os.path import exists
    from ..exceptions import NoInput
    from ..exceptions import ArgumentError
    from ..util.util import is_compatible_version
    from ..gui.util import get_host_port

    dbpath = args.get("result")
    if exists(dbpath) == False:
        raise NoInput()
    (compatible_version, db_version, oc_version) = is_compatible_version(dbpath)
    host, port = get_host_port(args=args)
    if not compatible_version:
        msg = f"DB version {db_version} of {dbpath} is not compatible with the current OakVar ({oc_version}). "
        msg += f'Consider running "oc util update-result {dbpath}" and running "oc gui {dbpath}" again.'
        raise ArgumentError(msg=msg)
    url = f"{host}:{port}/result/nocache/index.html?dbpath={args['result']}"
    return url


def get_login_url(args={}):
    from ..gui.util import get_host_port

    host, port = get_host_port(args=args)
    url = f"{host}:{port}/submit/nocache/login.html"
    return url


# def get_index_url(args={}):
#    from ..gui.util import get_host_port
#    host, port = get_host_port(args=args)
#    url = f"{host}:{port}/submit/nocache/index.html"
#    return url


def get_index_url(args={}):
    from ..gui.util import get_host_port

    host, port = get_host_port(args=args)
    url = f"{host}:{port}/index.html"
    return url


def get_url(args={}):
    if args.get("webapp"):
        url = get_webapp_url(args=args)
    elif args.get("result"):
        url = get_result_url(args=args)
    elif args.get("headless"):
        url = None
    else:
        if args.get("servermode"):
            url = get_login_url(args=args)
        else:
            url = get_index_url(args=args)
    if url:
        protocol = get_protocol(args=args)
        url = protocol + url
    return url


@cli_entry
def cli_gui(args):
    return gui(args)


@cli_func
def gui(args, __name__="gui"):
    from sys import stderr
    from traceback import print_exc
    from ..system import get_system_conf

    sysconf = get_system_conf()
    args["sysconf"] = sysconf
    args["logger"] = get_logger(args=args)
    try:
        setup(args=args)
        url = get_url(args=args)
        main(url=url, args=args)
    except Exception as e:
        # logger = args.get("logger")
        logger = get_logger(args=args)
        if logger:
            logger.exception(e)
        if args.get("debug"):
            print_exc()
        if logger:
            logger.info("Exiting...")
        stderr.write(f"{e}\nCheck {log_path} for details.\n")
    finally:
        logger = args.get("logger")
        if logger:
            for handler in logger.handlers:
                handler.close()
                logger.removeHandler(handler)



from .cli.module import add_parser_ov_module
from .cli.util import get_parser_fn_util
from .cli.run import add_parser_ov_run
from .cli.gui import get_parser_fn_gui
from .cli.report import get_parser_fn_report
from .cli.new import get_parser_fn_new
from .cli.issue import get_parser_fn_issue
from .cli.version import get_parser_fn_version
from .cli.store import get_parser_fn_store
from .cli.system import add_parser_ov_system
from .cli.config import get_parser_fn_config


def get_entry_parser():
    # subparsers
    from argparse import ArgumentParser

    p_entry = ArgumentParser(
        description="OakVar. Genomic variant analysis platform. https://github.com/rkimoakbioinformatics/oakvar"
    )
    p_entry.add_argument("--to", default="stdout")
    subparsers = p_entry.add_subparsers(title="Commands")
    # run
    add_parser_ov_run(subparsers)
    # report
    p_report = subparsers.add_parser(
        "report",
        parents=[get_parser_fn_report()],
        description="Generate reports from a job",
        add_help=False,
        help="Generate a report from a job",
        epilog="dbpath must be the first argument",
    )
    p_report.r_return = "A string, a named list, or a dataframe. Output of reporters"  # type: ignore
    p_report.r_examples = [  # type: ignore
        "# Generate a CSV-format report file from the job result file example.sqlite",
        '#roakvar::report(dbpath="example.sqlite", reports="csv")',
    ]

    # module
    add_parser_ov_module(subparsers)
    # gui
    p_gui = subparsers.add_parser(
        "gui", parents=[get_parser_fn_gui()], add_help=False, help="Start the GUI"
    )
    p_gui.r_return = "`NULL`"  # type: ignore
    p_gui.r_examples = [  # type: ignore
        "# Launch OakVar GUI",
        "#roakvar::gui()",
        "# Launch OakVar Interactive Result Viewer for the OakVar analysis file example.sqlite",
        '#roakvar::gui(result="example.sqlite")',
    ]

    # config
    _ = subparsers.add_parser(
        "config",
        parents=[get_parser_fn_config()],
        description="Manages OakVar configurations",
        add_help=False,
        help="Manages OakVar configurations",
    )
    # new
    _ = subparsers.add_parser(
        "new",
        parents=[get_parser_fn_new()],
        description="Create new modules",
        add_help=False,
        help="Create OakVar example input files and module templates",
    )
    # store
    _ = subparsers.add_parser(
        "store",
        parents=[get_parser_fn_store()],
        description="Publish modules to the store",
        add_help=False,
        help="Publish modules to the store",
    )
    # util
    _ = subparsers.add_parser(
        "util",
        parents=[get_parser_fn_util()],
        description="Utilities",
        add_help=False,
        help="OakVar utilities",
    )
    # version
    p_version = subparsers.add_parser(
        "version",
        parents=[get_parser_fn_version()],
        add_help=False,
        help="Show version",
    )
    p_version.r_return = "A string. OakVar version"  # type: ignore
    p_version.r_examples = [  # type: ignore
        "# Get the version of the installed OakVar",
        "#roakvar::version()",
    ]

    # issue
    p_issue = subparsers.add_parser(
        name="issue",
        parents=[get_parser_fn_issue()],
        add_help=False,
        help="Send an issue report",
    )
    p_issue.r_return = "`NULL`"  # type: ignore
    p_issue.r_examples = [  # type: ignore
        "# Open the Issues page of the OakVar GitHub website",
        "#roakvar::issue()",
    ]

    # system
    add_parser_ov_system(subparsers)
    """_ = subparsers.add_parser(
    )"""
    return p_entry


def handle_exception(e: Exception):
    from sys import stderr
    from traceback import print_exc
    from .exceptions import ExpectedException
    import sys

    msg = getattr(e, "msg", None)
    if msg:
        stderr.write(msg)
        stderr.write("\n")
        stderr.flush()
    trc = getattr(e, "traceback", None)
    if trc:
        print_exc()
    isatty = hasattr(sys, "ps1")  # interactive shell?
    halt = getattr(e, "halt", False)
    returncode = getattr(e, "returncode", 1)
    if hasattr(e, "traceback") and getattr(e, "traceback"):
        import traceback
        traceback.print_exc()
    if isinstance(e, ExpectedException):
        if halt:
            if isatty:
                return False
            else:
                exit(returncode)
        else:
            if isatty:
                return False
            else:
                return False
    elif isinstance(e, KeyboardInterrupt):
        pass
    else:
        raise e


def main():
    from sys import argv
    import signal
    from . import raise_break

    signal.signal(signal.SIGINT, raise_break)
    global get_entry_parser
    try:
        p_entry = get_entry_parser()
        args = p_entry.parse_args()
        if hasattr(args, "func"):
            ret = args.func(args)
            if getattr(args, "to", "return") != "stdout":
                return ret
            return True
        else:
            p_entry.parse_args(argv[1:] + ["--help"])
    except Exception as e:
        handle_exception(e)


if __name__ == "__main__":
    main()



from typing import Any
from typing import Optional
from typing import List
from typing import Dict
from typing import Tuple
from typing import TextIO
from oakvar import BaseConverter

STDIN = "stdin"


class MasterConverter(object):

    ALREADYCRV = 2

    def __init__(self, *inargs, **inkwargs):
        from re import compile
        from oakvar import get_wgs_reader

        self.logger = None
        self.serveradmindb = None
        self.crv_writer = None
        self.crs_writer = None
        self.crm_writer = None
        self.crl_writer = None
        self.converters = {}
        self.available_input_formats = []
        self.pipeinput = False
        self.input_paths = None
        self.input_dir = None
        self.input_path_dict = {}
        self.input_path_dict2 = {}
        self.input_file_handles: Dict[str, TextIO] = {}
        self.output_dir: Optional[str] = None
        self.output_base_fname: Optional[str] = None
        self.conf = None
        self.unique_variants = None
        self.error_logger = None
        self.unique_excs: Dict[str, int] = {}
        self.err_holder = []
        self.wpath = None
        self.err_path = None
        self.crm_path = None
        self.crs_path = None
        self.crl_path = None
        self.do_liftover = None
        self.do_liftover_chrM = None
        self.lifter = None
        self.module_options = None
        self.given_input_assembly: Optional[str] = None
        self.converter_by_input_path: Dict[str, Optional[BaseConverter]] = {}
        self.extra_output_columns = []
        self.file_num_valid_variants = 0
        self.file_error_lines = 0
        self.total_num_valid_variants = 0
        self.total_error_lines = 0
        self.fileno = 0
        self.total_num_converted_variants = 0
        self.input_formats: List[str] = []
        self.genome_assemblies: List[str] = []
        self.base_re = compile("^[ATGC]+|[-]+$")
        self.chromdict = {
            "chrx": "chrX",
            "chry": "chrY",
            "chrMT": "chrM",
            "chrMt": "chrM",
            "chr23": "chrX",
            "chr24": "chrY",
        }
        self.parse_cmd_args(inargs, inkwargs)
        self.setup_logger()
        self.wgsreader = get_wgs_reader(assembly="hg38")

    def get_genome_assembly(self, converter) -> str:
        from oakvar.system.consts import default_assembly_key
        from oakvar.exceptions import NoGenomeException
        from oakvar.system import get_user_conf

        if self.given_input_assembly:
            return self.given_input_assembly
        input_assembly = getattr(converter, "input_assembly", None)
        if input_assembly:
            return input_assembly
        user_conf = get_user_conf() or {}
        genome_assembly = user_conf.get(default_assembly_key, None)
        if genome_assembly:
            return genome_assembly
        raise NoGenomeException()

    def set_do_liftover(self, genome_assembly, converter, f):
        self.do_liftover = genome_assembly != "hg38"
        if hasattr(converter, "get_do_liftover_chrM"):
            self.do_liftover_chrM = converter.get_do_liftover_chrM(
                genome_assembly, f, self.do_liftover
            )
        else:
            self.do_liftover_chrM = self.do_liftover
        if self.logger:
            self.logger.info(f"liftover needed: {self.do_liftover}")
            self.logger.info(f"liftover for chrM needed: {self.do_liftover_chrM}")

    def setup_lifter(self, genome_assembly):
        from oakvar.util.admin_util import get_liftover_chain_paths
        from oakvar.exceptions import InvalidGenomeAssembly
        from pyliftover import LiftOver

        liftover_chain_paths = get_liftover_chain_paths()
        if genome_assembly not in liftover_chain_paths:
            raise InvalidGenomeAssembly(genome_assembly)
        self.lifter = LiftOver(liftover_chain_paths[genome_assembly])

    def parse_inputs(self, args):
        from pathlib import Path

        if not args["inputs"]:
            raise
        self.pipeinput = False
        if (
            args["inputs"] is not None
            and len(args["inputs"]) == 1
            and args["inputs"][0] == "-"
        ):
            self.pipeinput = True
        self.input_paths = []
        if self.pipeinput == False:
            self.input_paths = [str(Path(x).absolute()) for x in args["inputs"] if x != "-"]
        else:
            self.input_paths = [f"./{STDIN}"]
        self.input_dir = str(Path(self.input_paths[0]).parent)
        if self.pipeinput == False:
            for i in range(len(self.input_paths)):
                self.input_path_dict[i] = self.input_paths[i]
                self.input_path_dict2[self.input_paths[i]] = i
        else:
            self.input_path_dict[0] = self.input_paths[0]
            self.input_path_dict2[STDIN] = 0

    def parse_output_dir(self, args):
        from pathlib import Path
        from os import makedirs

        self.output_dir = None
        self.output_dir = args.get("output_dir") or None
        if not self.output_dir:
            self.output_dir = self.input_dir
        if not self.output_dir:
            raise
        if not (Path(self.output_dir).exists()):
            makedirs(self.output_dir)
        self.output_base_fname: Optional[str] = args.get("name")
        if not self.output_base_fname:
            if not self.input_paths:
                raise
            self.output_base_fname = Path(self.input_paths[0]).name

    def parse_cmd_args(self, inargs, inkwargs):
        from oakvar.exceptions import ExpectedException
        from oakvar.util.util import get_args
        from oakvar.util.run import get_module_options

        parser = self.get_cmd_args_parser()
        args = get_args(parser, inargs, inkwargs)
        if "serveradmindb" in args:
            self.serveradmindb = args.get("serveradmindb")
        if args["inputs"] is None:
            raise ExpectedException("Input files are not given.")
        self.parse_inputs(args)
        self.parse_output_dir(args)
        self.given_input_assembly = args.get("genome")
        self.conf = {}
        self.module_options = get_module_options(args)
        if "conf" in args:
            self.conf.update(args["conf"])
        self.unique_variants = args["unique_variants"]
        self.args = args

    def get_file_object_for_input_path(self, input_path: str):
        import gzip
        from oakvar.util.util import detect_encoding

        if self.logger:
            self.logger.info(f"detecting encoding of {input_path}")
        encoding = detect_encoding(input_path)
        if self.logger:
            self.logger.info(f"encoding: {input_path} {encoding}")
        if input_path.endswith(".gz"):
            f = gzip.open(input_path, mode="rt", encoding=encoding)
        else:
            f = open(input_path, encoding=encoding)
        return f

    def collect_input_file_handles(self):
        if not self.input_paths:
            raise
        for input_path in self.input_paths:
            f = self.get_file_object_for_input_path(input_path)
            self.input_file_handles[input_path] = f

    def setup(self):
        self.collect_input_file_handles()
        self.collect_converters()
        self.check_input_format()
        self.collect_converter_by_input()
        self.collect_extra_output_columns()
        self.open_output_files()

    def setup_logger(self):
        from logging import getLogger

        self.logger = getLogger("oakvar.converter")
        self.error_logger = getLogger("err.converter")

    def collect_converters(self):
        from oakvar.module.local import get_local_module_infos_of_type
        from oakvar.util.util import load_class
        from oakvar.util.util import quiet_print

        for module_info in get_local_module_infos_of_type("converter").values():
            cls = load_class(module_info.script_path)
            converter = cls()
            # TODO: backward compatibility
            converter.output_dir = None
            converter.run_name = None
            converter.module_name = module_info.name
            converter.name = module_info.name
            converter.version = module_info.version
            converter.conf = module_info.conf
            converter.script_path = module_info.script_path
            # end of backward compatibility
            if not hasattr(converter, "format_name"):
                quiet_print("{module_info.name} does not have format_name defined and thus was skipped.", args=self.args)
                continue
            converter.module_name = module_info.name
            if converter.format_name not in self.converters:
                self.converters[converter.format_name] = converter
            else:
                quiet_print("{moule_info.name} is skipped because {converter.format_name} is already handled by {self.converters[converter.format_name].name}.", args=self.args)
                continue
        self.available_input_formats = list(self.converters.keys())

    def collect_converter_by_input(self):
        if not self.input_paths:
            return
        for input_path in self.input_paths:
            f = self.input_file_handles[input_path]
            converter = self.get_converter_for_input_file(f)
            self.converter_by_input_path[input_path] = converter

    def collect_extra_output_columns(self):
        for converter in self.converter_by_input_path.values():
            if not converter:
                continue
            if converter.conf:
                extra_output_columns = converter.conf.get("extra_output_columns")
                if not extra_output_columns:
                    continue
                for col in extra_output_columns:
                    self.extra_output_columns.append(col)

    def check_input_format(self):
        from oakvar.exceptions import InvalidInputFormat

        if self.args.get("input_format") and self.args.get("input_format") not in self.available_input_formats:
            raise InvalidInputFormat(self.args.get("input_format"))
        if self.pipeinput and not self.args.get("input_format"):
            raise InvalidInputFormat(
                f"--input-format should be given with pipe input"
            )

    def get_converter_for_input_file(self, f) -> Optional[BaseConverter]:
        import sys
        from oakvar.exceptions import ArgumentError

        if f == sys.stdin:
            if not self.args.get("input_format"):
                raise ArgumentError(msg="--input-format option should be given if stdin is used for input.")
            if not self.args.get("input_format") in self.converters:
                raise ArgumentError(msg=f"{self.args.get('input_format')} is not found in installed converter modules.")
            return self.converters[self.args.get("input_format")]
        else:
            for converter_name, converter in self.converters.items():
                f.seek(0)
                try:
                    check_success = converter.check_format(f)
                except:
                    import traceback
                    traceback.print_exc()
                    check_success = False
                f.seek(0)
                if check_success:
                    if self.logger:
                        self.logger.info(f"using {converter_name} for {f.name}")
                    return converter
        return None

    def set_converter_properties(self, converter):
        from oakvar.exceptions import SetupError
        from oakvar.module.local import get_module_code_version

        if self.conf is None:
            raise SetupError()
        converter.output_dir = self.output_dir
        converter.run_name = self.output_base_fname
        module_name = converter.format_name + "-converter"
        converter.module_name = module_name
        converter.version = get_module_code_version(converter.module_name)
        if module_name in self.conf:
            if hasattr(converter, "conf") == False:
                converter.conf = {}
            converter.conf.update(self.conf[module_name])

    def setup_crv_writer(self):
        from pathlib import Path
        from oakvar.util.util import get_crv_def
        from oakvar.util.inout import FileWriter
        from oakvar.consts import crv_idx
        from oakvar.consts import STANDARD_INPUT_FILE_SUFFIX

        if not self.output_dir or not self.output_base_fname:
            raise
        crv_def = get_crv_def()
        self.wpath = Path(self.output_dir) / (self.output_base_fname + STANDARD_INPUT_FILE_SUFFIX)
        self.crv_writer = FileWriter(self.wpath)
        self.crv_writer.add_columns(crv_def)
        self.crv_writer.write_definition()
        for index_columns in crv_idx:
            self.crv_writer.add_index(index_columns)

    def setup_crs_writer(self):
        from pathlib import Path
        from oakvar.util.util import get_crs_def
        from oakvar.util.inout import FileWriter
        from oakvar.consts import crs_idx
        from oakvar.consts import SAMPLE_FILE_SUFFIX

        if not self.output_dir or not self.output_base_fname:
            raise
        crs_def = get_crs_def()
        self.crs_path = Path(self.output_dir) / (self.output_base_fname + SAMPLE_FILE_SUFFIX)
        self.crs_writer = FileWriter(self.crs_path)
        self.crs_writer.add_columns(crs_def)
        self.crs_writer.add_columns(self.extra_output_columns)
        self.crs_writer.write_definition()
        for index_columns in crs_idx:
            self.crs_writer.add_index(index_columns)

    def setup_crm_writer(self):
        from pathlib import Path
        from oakvar.util.util import get_crm_def
        from oakvar.util.inout import FileWriter
        from oakvar.consts import crm_idx
        from oakvar.consts import MAPPING_FILE_SUFFIX

        if not self.output_dir or not self.output_base_fname:
            raise
        crm_def = get_crm_def()
        self.crm_path = Path(self.output_dir) / (self.output_base_fname + MAPPING_FILE_SUFFIX)
        self.crm_writer = FileWriter(self.crm_path)
        self.crm_writer.add_columns(crm_def)
        self.crm_writer.write_definition()
        for index_columns in crm_idx:
            self.crm_writer.add_index(index_columns)
        self.crm_writer.write_input_paths(self.input_path_dict)

    def setup_crl_writer(self):
        from pathlib import Path
        from oakvar.util.util import get_crl_def
        from oakvar.util.inout import FileWriter
        from oakvar.consts import VARIANT_LEVEL_OUTPUT_SUFFIX

        if not self.output_dir or not self.output_base_fname:
            raise
        crl_def = get_crl_def()
        self.crl_path = Path(self.output_dir) / f"{self.output_base_fname}.original_input{VARIANT_LEVEL_OUTPUT_SUFFIX}"
        self.crl_writer = FileWriter(self.crl_path)
        self.crl_writer.add_columns(crl_def)
        self.crl_writer.write_definition()
        self.crl_writer.write_names("original_input", "Original Input", "")

    def open_output_files(self):
        from pathlib import Path
        from oakvar.exceptions import SetupError

        if not self.output_base_fname or not self.output_dir:
            raise SetupError()
        self.err_path = Path(self.output_dir) / (self.output_base_fname + ".master_converter.err")
        self.setup_crv_writer()
        self.setup_crs_writer()
        self.setup_crm_writer()
        self.setup_crl_writer()

    def log_input_and_genome_assembly(self, input_path, genome_assembly, converter):
        if not self.logger:
            return
        if self.pipeinput:
            self.logger.info(f"input file: {STDIN}")
        else:
            self.logger.info(f"input file: {input_path}")
        self.logger.info("input format: %s" % converter.format_name)
        self.logger.info(f"genome_assembly: {genome_assembly}")

    def setup_file(self, input_path: str) -> Tuple[TextIO, BaseConverter]:
        from sys import stdin
        from logging import getLogger
        from oakvar.util.util import log_module
        from oakvar.exceptions import NoConverterFound

        if self.pipeinput:
            f = stdin
        else:
            f = self.input_file_handles[input_path]
        converter = self.converter_by_input_path[input_path]
        if not converter:
            raise NoConverterFound(input_path)
        if not converter.module_name or not converter.format_name:
            raise NoConverterFound(input_path)
        self.input_formats.append(converter.format_name)
        self.fileno += 1
        self.set_converter_properties(converter)
        log_module(converter, self.logger)
        self.error_logger = getLogger("err." + converter.module_name)
        converter.input_path = input_path
        converter.input_paths = self.input_paths
        converter.setup(f)
        genome_assembly = self.get_genome_assembly(converter)
        self.genome_assemblies.append(genome_assembly)
        self.log_input_and_genome_assembly(input_path, genome_assembly, converter)
        self.set_do_liftover(genome_assembly, converter, f)
        if self.do_liftover or self.do_liftover_chrM:
            self.setup_lifter(genome_assembly)
        if self.pipeinput == False:
            f.seek(0)
        return (f, converter)

    def handle_chrom(self, variant):
        from oakvar.exceptions import IgnoredVariant

        if not variant.get("chrom"):
            raise IgnoredVariant("No chromosome")
        if not variant.get("chrom").startswith("chr"):
            variant["chrom"] = "chr" + variant.get("chrom")
        variant["chrom"] = self.chromdict.get(variant.get("chrom"), variant.get("chrom"))

    def handle_ref_base(self, variant):
        from oakvar.exceptions import IgnoredVariant

        if "ref_base" not in variant or variant["ref_base"] in [
            "",
            ".",
        ]:
            if not self.wgsreader:
                raise
            variant["ref_base"] = self.wgsreader.get_bases(
                variant.get("chrom"), int(variant["pos"])
            ).upper()
        else:
            ref_base = variant["ref_base"]
            if ref_base == "" and variant["alt_base"] not in [
                "A",
                "T",
                "C",
                "G",
            ]:
                raise IgnoredVariant(
                    "Reference base required for non SNV"
                )
            elif ref_base is None or ref_base == "":
                if not self.wgsreader:
                    raise
                variant["ref_base"] = self.wgsreader.get_bases(
                    variant.get("chrom"), int(variant.get("pos"))
                )

    def handle_genotype(self, variant):
        if "genotype" in variant and "." in variant["genotype"]:
            variant["genotype"] = variant["genotype"].replace(
                ".", variant["ref_base"]
            )

    def check_invalid_base(self, variant: dict):
        from oakvar.exceptions import IgnoredVariant

        if not self.base_re.fullmatch(variant["ref_base"]):
            raise IgnoredVariant("Invalid reference base")
        if not self.base_re.fullmatch(variant["alt_base"]):
            raise IgnoredVariant("Invalid alternate base")

    def normalize_variant(self, variant):
        from oakvar.util.seq import normalize_variant_left

        p, r, a = (
            int(variant["pos"]),
            variant["ref_base"],
            variant["alt_base"],
        )
        (
            new_pos,
            new_ref,
            new_alt,
        ) = normalize_variant_left("+", p, r, a)
        variant["pos"] = new_pos
        variant["ref_base"] = new_ref
        variant["alt_base"] = new_alt
        variant["uid"] = self.uid

    def handle_variant(self, variant: dict, var_no: int, converter):
        from oakvar.exceptions import NoVariantError

        if not self.crv_writer or not self.crm_writer or not self.crs_writer:
            raise
        self.handle_chrom(variant)
        self.handle_ref_base(variant)
        self.handle_genotype(variant)
        self.perform_liftover_if_needed(variant)
        self.check_invalid_base(variant)
        self.normalize_variant(variant)
        if variant["ref_base"] == variant["alt_base"]:
            raise NoVariantError()
        self.file_num_valid_variants += 1
        self.total_num_valid_variants += 1
        self.crv_writer.write_data(variant)
        try:
            converter.addl_operation_for_unique_variant(
                variant, var_no
            )
        except Exception as e:
            self._log_conversion_error(
                self.read_lnum, e, full_line_error=False
            )
        self.crm_writer.write_data(
            {
                "original_line": self.read_lnum,
                "tags": variant["tags"],
                "uid": self.uid,
                "fileno": f"{self.input_path_dict2[self.input_path]}",
            }
        )
        self.crs_writer.write_data(variant)

    def handle_converted_variants(self, variants: List[Dict[str, Any]], converter: BaseConverter):
        from oakvar.exceptions import IgnoredVariant

        if not self.wgsreader or not self.crv_writer or not self.crm_writer or not self.crs_writer:
            raise
        if variants is BaseConverter.IGNORE:
            return
        self.total_num_converted_variants += 1
        if not variants:
            raise IgnoredVariant(
                "No valid alternate allele was found in any samples."
            )
        for var_no, variant in enumerate(variants):
            self.handle_variant(variant, var_no, converter)
            self.uid += 1

    def run(self):
        from pathlib import Path
        from oakvar.util.run import update_status

        if not self.input_paths or not self.logger:
            raise
        update_status(f"started converter", logger=self.logger, serveradmindb=self.serveradmindb)
        self.setup()
        self.set_variables_pre_run()
        for self.input_path in self.input_paths:
            self.input_fname = Path(self.input_path).name if not self.pipeinput else STDIN
            f, converter = self.setup_file(self.input_path)
            self.file_num_valid_variants = 0
            self.file_error_lines = 0
            for self.read_lnum, variants in converter.convert_file(
                f, exc_handler=self._log_conversion_error
            ):
                try:
                    self.handle_converted_variants(variants, converter)
                    if self.read_lnum % 10000 == 0:
                        status = f"Running Converter ({self.input_fname}): line {self.read_lnum}"
                        update_status(
                            status, logger=self.logger, serveradmindb=self.serveradmindb
                        )
                except Exception as e:
                    self._log_conversion_error(self.read_lnum, e)
            f.close()
            self.logger.info(f"number of valid variants: {self.file_num_valid_variants}")
            self.logger.info(f"number of error lines: {self.file_error_lines}")
        self.close_output_files()
        self.end()
        self.log_ending()
        ret = {
            "total_lnum": self.total_num_converted_variants,
            "write_lnum": self.total_num_valid_variants,
            "error_lnum": self.total_error_lines,
            "format": self.input_formats,
            "assemblies": self.genome_assemblies,
        }
        return ret

    def set_variables_pre_run(self):
        from time import time

        self.start_time = time()
        self.total_num_converted_variants = 0
        self.uid = 1

    def log_ending(self):
        from time import time, asctime, localtime
        from oakvar.util.run import update_status

        if not self.logger:
            raise
        self.logger.info("total number of converted variants: {}".format(self.total_num_converted_variants))
        self.logger.info("number of total error lines: %d" % self.total_error_lines)
        end_time = time()
        self.logger.info("finished: %s" % asctime(localtime(end_time)))
        runtime = round(end_time - self.start_time, 3)
        self.logger.info("runtime: %s" % runtime)
        status = f"finished Converter"
        update_status(status, logger=self.logger, serveradmindb=self.serveradmindb)

    def perform_liftover_if_needed(self, variant):
        from copy import copy

        if not self.crl_writer:
            raise
        prelift_wdict = copy(variant)
        prelift_wdict["uid"] = self.uid
        self.crl_writer.write_data(prelift_wdict)
        if self.is_chrM(variant):
            needed = self.do_liftover_chrM
        else:
            needed = self.do_liftover
        if needed:
            (
                variant["chrom"],
                variant["pos"],
                variant["ref_base"],
                variant["alt_base"],
            ) = self.liftover(
                variant["chrom"],
                int(variant["pos"]),
                variant["ref_base"],
                variant["alt_base"],
            )

    def is_chrM(self, wdict):
        return wdict["chrom"] == "chrM"

    def liftover_one_pos(self, chrom, pos):
        from oakvar.exceptions import SetupError

        if not self.lifter:
            raise SetupError("no lifter")
        res = self.lifter.convert_coordinate(chrom, pos - 1)
        if res is None or len(res) == 0:
            res_prev = self.lifter.convert_coordinate(chrom, pos - 2)
            res_next = self.lifter.convert_coordinate(chrom, pos)
            if res_prev is not None and res_next is not None:
                if len(res_prev) == 1 and len(res_next) == 1:
                    pos_prev = res_prev[0][1]
                    pos_next = res_next[0][1]
                    if pos_prev == pos_next - 2:
                        res = [(res_prev[0][0], pos_prev + 1)]
                    elif pos_prev == pos_next + 2:
                        res = [(res_prev[0][0], pos_prev - 1)]
        return res

    def liftover(self, chrom, pos, ref, alt):
        from oakvar.exceptions import LiftoverFailure
        from oakvar.util.seq import reverse_complement
        from oakvar.exceptions import SetupError

        if not self.lifter or not self.wgsreader:
            raise SetupError("no lifter")
        reflen = len(ref)
        altlen = len(alt)
        if reflen == 1 and altlen == 1:
            res = self.liftover_one_pos(chrom, pos)
            if res is None or len(res) == 0:
                raise LiftoverFailure("Liftover failure")
            if len(res) > 1:
                raise LiftoverFailure("Liftover failure")
            try:
                el = res[0]
            except:
                raise LiftoverFailure("Liftover failure")
            newchrom = el[0]
            newpos = el[1] + 1
        elif reflen >= 1 and altlen == 0:  # del
            pos1 = pos
            pos2 = pos + reflen - 1
            res1 = self.lifter.convert_coordinate(chrom, pos1 - 1)
            res2 = self.lifter.convert_coordinate(chrom, pos2 - 1)
            if res1 is None or res2 is None or len(res1) == 0 or len(res2) == 0:
                raise LiftoverFailure("Liftover failure")
            if len(res1) > 1 or len(res2) > 1:
                raise LiftoverFailure("Liftover failure")
            el1 = res1[0]
            el2 = res2[0]
            newchrom1 = el1[0]
            newpos1 = el1[1] + 1
            newpos2 = el2[1] + 1
            newchrom = newchrom1
            newpos = newpos1
            newpos = min(newpos1, newpos2)
        elif reflen == 0 and altlen >= 1:  # ins
            res = self.lifter.convert_coordinate(chrom, pos - 1)
            if res is None or len(res) == 0:
                raise LiftoverFailure("Liftover failure")
            if len(res) > 1:
                raise LiftoverFailure("Liftover failure")
            el = res[0]
            newchrom = el[0]
            newpos = el[1] + 1
        else:
            pos1 = pos
            pos2 = pos + reflen - 1
            res1 = self.lifter.convert_coordinate(chrom, pos1 - 1)
            res2 = self.lifter.convert_coordinate(chrom, pos2 - 1)
            if res1 is None or res2 is None or len(res1) == 0 or len(res2) == 0:
                raise LiftoverFailure("Liftover failure")
            if len(res1) > 1 or len(res2) > 1:
                raise LiftoverFailure("Liftover failure")
            el1 = res1[0]
            el2 = res2[0]
            newchrom1 = el1[0]
            newpos1 = el1[1] + 1
            newpos2 = el2[1] + 1
            newchrom = newchrom1
            newpos = min(newpos1, newpos2)
        hg38_ref = self.wgsreader.get_bases(newchrom, newpos)
        if hg38_ref == reverse_complement(ref):
            newref = hg38_ref
            newalt = reverse_complement(alt)
        else:
            newref = ref
            newalt = alt
        return [newchrom, newpos, newref, newalt]

    def _log_conversion_error(self, line_no: int, e, full_line_error=True):
        from traceback import format_exc
        from oakvar.exceptions import ExpectedException
        from oakvar.exceptions import NoAlternateAllele
        from oakvar.util.run import update_status

        if isinstance(e, NoAlternateAllele):
            if line_no % 10000 == 0:
                status = f"Running Converter ({self.input_fname}): line {line_no}"
                update_status(
                    status, logger=self.logger, serveradmindb=self.serveradmindb
                )
            return
        if not self.logger or not self.error_logger:
            raise
        if full_line_error:
            self.file_error_lines += 1
            self.total_error_lines += 1
        if isinstance(e, ExpectedException):
            err_str = str(e)
        else:
            err_str = format_exc().rstrip()
        if err_str not in self.unique_excs:
            err_no = len(self.unique_excs)
            self.unique_excs[err_str] = err_no
            if hasattr(e, "traceback") and e.traceback == False:
                pass
            else:
                self.logger.error(f"Error [{err_no}]: {self.input_path}: {err_str}")
            self.err_holder.append(f"{err_no}:{line_no}\t{str(e)}")
        else:
            err_no = self.unique_excs[err_str]
            self.err_holder.append(f"{err_no}:{line_no}\t{str(e)}")
        if len(self.err_holder) % 1000 == 0:
            for s in self.err_holder:
                self.error_logger.error(s)
            self.err_holder = []
        if line_no % 10000 == 0:
            status = f"Running Converter ({self.input_fname}): line {line_no}"
            update_status(
                status, logger=self.logger, serveradmindb=self.serveradmindb
            )

    def close_output_files(self):
        if self.crv_writer is not None:
            self.crv_writer.close()
        if self.crm_writer is not None:
            self.crm_writer.close()
        if self.crs_writer is not None:
            self.crs_writer.close()

    def end(self):
        pass


    def get_cmd_args_parser(self):
        from argparse import ArgumentParser, SUPPRESS

        parser = ArgumentParser()
        parser.add_argument("path", help="Path to this converter's python module")
        parser.add_argument(
            "inputs", nargs="*", default=None, help="Files to be converted to .crv"
        )
        parser.add_argument(
            "-f", dest="format", default=None, help="Specify an input format"
        )
        parser.add_argument(
            "-n", "--name", dest="name", help="Name of job. Default is input file name."
        )
        parser.add_argument(
            "-d",
            "--output-dir",
            dest="output_dir",
            help="Output directory. Default is input file directory.",
        )
        parser.add_argument(
            "-l",
            "--genome",
            dest="genome",
            default=None,
            help="Input gene assembly. Will be lifted over to hg38",
        )
        parser.add_argument(
            "--confs", dest="confs", default="{}", help="Configuration string"
        )
        parser.add_argument(
            "--unique-variants",
            dest="unique_variants",
            default=False,
            action="store_true",
            help=SUPPRESS,
        )
        return parser


def main():
    master_converter = MasterConverter()
    master_converter.run()


if __name__ == "__main__":
    master_converter = MasterConverter()
    master_converter.run()



from typing import Optional
from typing import List
from typing import Tuple
from typing import Any


def get_ov_store_cache_conn(conf=None):
    from sqlite3 import connect
    from .consts import ov_store_cache_fn
    from os.path import join
    from ..system import get_conf_dir

    conf_dir: Optional[str] = get_conf_dir(conf=conf)
    if conf_dir:
        ov_store_cache_path = join(conf_dir, ov_store_cache_fn)
        conn = connect(ov_store_cache_path)
        cursor = conn.cursor()
        return conn, cursor
    return None, None


def db_func(func):
    def encl_func(*args, conf=None, **kwargs):
        conn, cursor = get_ov_store_cache_conn(conf=conf)
        ret = func(*args, conn=conn, cursor=cursor, **kwargs)
        return ret

    return encl_func


@db_func
def find_name_store(
    module_name: str, conn=None, cursor=None
) -> Optional[Tuple[str, str]]:
    if not conn or not cursor:
        return None
    q = f"select name, store from summary where name=?"
    cursor.execute(q, (module_name,))
    ret = cursor.fetchall()
    name = None
    store = None
    for r in ret:
        if not name or r[1] == "ov":
            name = r[0]
            store = r[1]
    if name and store:
        return name, store
    else:
        return None


@db_func
def latest_module_version_size(
    module_name: str, conn=Any, cursor=Any
) -> Optional[dict]:
    from packaging.version import Version

    _ = conn
    q = f"select store, code_version, data_version, data_source, code_size, data_size from versions where name=?"
    cursor.execute(q, (module_name,))
    ret = cursor.fetchall()
    latest_code_version = ""
    latest_r = None
    for r in ret:
        if not latest_r:
            latest_r = r
            latest_code_version = r[1]
        elif Version(r[1]) > Version(latest_code_version):
            latest_r = r
            latest_code_version = r[1]
    if latest_r:
        return {
            "code_version": latest_r[1],
            "data_version": latest_r[2],
            "data_source": latest_r[3],
            "code_size": int(latest_r[4]),
            "data_size": int(latest_r[5]),
        }
    else:
        return None


@db_func
def module_code_versions(module_name, conn=None, cursor=None) -> Optional[List[str]]:
    if not conn or not cursor:
        return None
    r = find_name_store(module_name)
    if not r:
        return None
    name, store = r
    q = f"select code_version from versions where name=? and store=?"
    cursor.execute(q, (name, store))
    values = [r[0] for r in cursor.fetchall()]
    return values


@db_func
def module_data_versions(module_name, conn=None, cursor=None) -> Optional[List[str]]:
    if not conn or not cursor:
        return None
    r = find_name_store(module_name)
    if not r:
        return None
    name, store = r
    q = f"select data_version from versions where name=? and store=?"
    cursor.execute(q, (name, store))
    values = [r[0] for r in cursor.fetchall()]
    return values


@db_func
def module_data_sources(module_name, conn=Any, cursor=Any) -> Optional[List[str]]:
    _ = conn or cursor
    r = find_name_store(module_name)
    if not r:
        return None
    name, store = r
    q = f"select data_source from versions where name=? and store=?"
    cursor.execute(q, (name, store))
    values = [r[0] for r in cursor.fetchall()]
    return values


@db_func
def module_min_pkg_vers(module_name, conn=Any, cursor=Any) -> Optional[List[str]]:
    _ = conn or cursor
    r = find_name_store(module_name)
    if not r:
        return None
    name, store = r
    q = f"select min_pkg_ver from versions where name=? and store=?"
    cursor.execute(q, (name, store))
    values = [r[0] for r in cursor.fetchall()]
    return values


@db_func
def module_sizes(
    module_name: str, code_version: str, conn=None, cursor=None
) -> Optional[Tuple[int, int]]:
    if not conn or not cursor:
        return None
    r = find_name_store(module_name)
    if not r:
        return None
    name, store = r
    q = f"select code_size, data_size from versions where name=? and store=? and code_version=?"
    cursor.execute(q, (name, store, code_version))
    r = cursor.fetchone()
    if not r:
        return None
    code_size, data_size = r
    return int(code_size), int(data_size)


@db_func
def module_data_source(
    module_name: str, code_version: str, conn=None, cursor=None
) -> Optional[str]:
    if not conn or not cursor:
        return None
    r = find_name_store(module_name)
    if not r:
        return None
    name, store = r
    q = f"select data_source from versions where name=? and store=? and code_version=?"
    cursor.execute(q, (name, store, code_version))
    r = cursor.fetchone()
    if not r:
        return None
    return r[0]


@db_func
def remote_module_data_version(
    module_name: str, code_version: str, conn=None, cursor=None
) -> Optional[str]:
    if not conn or not cursor:
        return None
    r = find_name_store(module_name)
    if not r:
        return None
    name, store = r
    q = f"select data_version from versions where name=? and store=? and code_version=?"
    cursor.execute(q, (name, store, code_version))
    r = cursor.fetchone()
    if not r:
        return None
    return r[0]


@db_func
def get_latest_module_code_version(module_name, conn=None, cursor=None):
    from ..util.util import get_latest_version
    from ..util.admin_util import oakvar_version
    from packaging.version import Version

    if not conn or not cursor:
        return None
    r = find_name_store(module_name)
    if not r:
        return None
    pkg_ver = Version(oakvar_version())
    name, store = r
    q = f"select code_version, min_pkg_ver from versions where name=? and store=?"
    cursor.execute(q, (name, store))
    code_versions = []
    for row in cursor.fetchall():
        [code_version, min_pkg_ver] = row
        if min_pkg_ver.startswith(">="):
            min_pkg_ver = min_pkg_ver[2:]
        if pkg_ver >= Version(min_pkg_ver):
            code_versions.append(code_version)
    latest_code_version = get_latest_version(code_versions)
    return latest_code_version


@db_func
def module_code_version_is_not_compatible_with_pkg_version(module_name: str, code_version: str, conn=Any, cursor=Any) -> Optional[str]:
    from packaging.version import Version
    from ..util.admin_util import oakvar_version

    _ = conn
    pkg_ver = Version(oakvar_version())
    q = f"select min_pkg_ver from versions where name=? and code_version=?"
    cursor.execute(q, (module_name, code_version))
    min_pkg_ver = None
    for row in cursor.fetchall():
        version = row[0]
        if Version(version) <= pkg_ver:
            return None
        else:
            if min_pkg_ver is None or Version(version) < Version(min_pkg_ver):
                min_pkg_ver = version
    if min_pkg_ver is None:
        min_pkg_ver = "?"
    return min_pkg_ver

@db_func
def module_info_ls(module_name, conn=None, cursor=None):
    import sqlite3
    from .consts import summary_table_cols
    from ..module.remote import RemoteModuleLs

    if not conn or not cursor:
        return None
    cursor.row_factory = sqlite3.Row
    q = f"select { ', '.join(summary_table_cols) } from summary where name=?"
    cursor.execute(q, (module_name,))
    ret = cursor.fetchall()
    module_info = None
    for r in ret:
        info = {}
        for col in summary_table_cols:
            info[col] = r[col]
        if not module_info or info["store"] == "ov":
            module_info = RemoteModuleLs("", **info)
    return module_info


@db_func
def module_info(module_name, conn=None, cursor=None):
    import sqlite3
    from .consts import summary_table_cols
    from ..module.remote import RemoteModule

    if not conn or not cursor:
        return None
    cursor.row_factory = sqlite3.Row
    q = f"select { ', '.join(summary_table_cols) } from summary where name=?"
    cursor.execute(q, (module_name,))
    ret = cursor.fetchall()
    module_info = None
    for r in ret:
        info = {}
        for col in summary_table_cols:
            info[col] = r[col]
        if not module_info or info["store"] == "ov":
            module_info = RemoteModule("", **info)
    return module_info


@db_func
def summary_col_value(module_name: str, colname: str, conn=None, cursor=None):
    from json import loads

    if not conn or not cursor:
        return None
    q = f"select {colname}, store from summary where name=?"
    cursor.execute(q, (module_name,))
    ret = cursor.fetchall()
    out = None
    if ret:
        for r in ret:
            v, store = r
            if not out or store == "ov":
                out = v
    if out:
        if out[0] in ["[", "{"]:
            return loads(out)
        else:
            return out
    else:
        return None


@db_func
def module_list(module_type=None, conn=None, cursor=None) -> List[str]:
    if not conn or not cursor:
        return []
    if module_type:
        q = f"select distinct(name) from summary where type=?"
        cursor.execute(q, (module_type,))
    else:
        q = f"select distinct(name) from summary"
        cursor.execute(q)
    ret = cursor.fetchall()
    l = set()
    for v in ret:
        l.add(v[0])
    return list(l)


@db_func
def table_exists(table: str, conf=None, conn=None, cursor=None) -> bool:
    if not conn or not cursor:
        return False
    if conf:
        pass
    q = f"select name from sqlite_master where type='table' and name=?"
    cursor.execute(q, (table,))
    ret = cursor.fetchone()
    if not ret:
        return False
    else:
        return True


@db_func
def is_store_db_schema_changed(conn=Any, cursor=Any) -> bool:
    from .consts import summary_table_cols
    from .consts import versions_table_cols

    _ = conn
    q = f'pragma table_info("summary")'
    cursor.execute(q)
    cols = [row[1] for row in cursor.fetchall()]
    if len(cols) > 0 and cols != summary_table_cols:
        return True
    q = f'pragma table_info("versions")'
    cursor.execute(q)
    cols = [row[1] for row in cursor.fetchall()]
    if len(cols) > 0 and cols != versions_table_cols:
        return True
    return False

@db_func
def drop_ov_store_cache(conf=None, conn=None, cursor=None, args={}):
    from os.path import exists
    from ..system import get_cache_dir
    from ..system.consts import cache_dirs
    from shutil import rmtree

    if not conn or not cursor:
        return
    if conf:
        pass
    if args.get("clean_cache_db"):
        for table in ["summary", "versions", "info"]:
            if table_exists(table):
                q = f"drop table if exists {table}"
                cursor.execute(q)
                conn.commit()
    if args.get("clean_cache_files"):
        for cache_key in cache_dirs:
            fp = get_cache_dir(cache_key)
            if exists(fp):
                rmtree(fp)


@db_func
def create_ov_store_cache(conf=None, args={}, conn=None, cursor=None):
    from .consts import summary_table_cols
    from .consts import versions_table_cols
    from ..system.consts import cache_dirs
    from ..system import get_cache_dir
    from os.path import exists
    from os.path import join
    from os import mkdir

    if not conn or not cursor:
        return False
    if conf:
        pass
    clean_cache_db = args.get("clean_cache_db")
    if clean_cache_db or not table_exists("summary"):
        q = f"create table summary ( { ', '.join([col + ' text' for col in summary_table_cols]) }, primary key ( name, store ) )"
        cursor.execute(q)
    if clean_cache_db or not table_exists("versions"):
        q = f"create table versions ( { ', '.join([col + ' text' for col in versions_table_cols]) }, primary key ( name, store, code_version ) )"
        cursor.execute(q)
    if clean_cache_db or not table_exists("info"):
        q = f"create table info ( key text primary key, value text )"
        cursor.execute(q)
    conn.commit()
    for cache_key in cache_dirs:
        fp = get_cache_dir(cache_key)
        if not exists(fp):
            mkdir(fp)
            mkdir(join(fp, "ov"))
            mkdir(join(fp, "oc"))


def try_fetch_ov_store_cache(args={}):
    from ..util.util import quiet_print

    try:
        return fetch_ov_store_cache(args=args)
    except Exception as e:
        quiet_print(
            f"Fetching store update failed:\n\n>>{e}.\n\nContinuing with the current store cache...\n",
            args=args,
        )


@db_func
def fetch_ov_store_cache(
    conn=None,
    cursor=None,
    args={},
):
    from .consts import ov_store_last_updated_col
    from ..util.util import quiet_print
    from ..exceptions import StoreServerError
    from ..exceptions import AuthorizationError
    from .ov.account import login_with_token_set
    from .ov import get_server_last_updated
    from ..module.remote import make_remote_manifest
    from ..gui.webstore.webstore import save_remote_manifest_cache

    if not conn or not cursor:
        return False
    if not login_with_token_set():
        quiet_print(f"not logged in", args=args)
        return False
    if is_new_store_db_setup():
        args["clean_cache_db"] = True
        args["clean_cache_files"] = True
        args["clean"] = True
        local_last_updated = None
    else:
        local_last_updated = get_local_last_updated()
    if is_store_db_schema_changed():
        quiet_print(f"Need to fetch store cache due to schema change", args=args)
        args["clean_cache_db"] = True
    clean_cache_files = args.get("clean_cache_files")
    clean_cache_db = args.get("clean_cache_db")
    server_last_updated, status_code = get_server_last_updated()
    if not server_last_updated:
        if status_code == 401:
            raise AuthorizationError()
        elif status_code == 500:
            raise StoreServerError()
        return False
    if (
        not clean_cache_db
        and not clean_cache_files
        and local_last_updated
        and local_last_updated >= server_last_updated
    ):
        quiet_print("No store update to fetch", args=args)
        return True
    args["publish_time"] = local_last_updated
    drop_ov_store_cache(args=args)
    create_ov_store_cache(args=args)
    fetch_summary_cache(args=args)
    fetch_versions_cache(args=args)
    if args.get("clean_cache_files") or args.get("clean"):
        args["publish_time"] = ""
    else:
        args["publish_time"] = local_last_updated
    fetch_readme_cache(args=args)
    fetch_logo_cache(args=args)
    fetch_conf_cache(args=args)
    q = f"insert or replace into info ( key, value ) values ( ?, ? )"
    cursor.execute(q, (ov_store_last_updated_col, str(server_last_updated)))
    conn.commit()
    content = make_remote_manifest()
    save_remote_manifest_cache(content)
    quiet_print("OakVar store cache has been fetched.", args=args)
    return True


@db_func
def is_new_store_db_setup(conn=Any, cursor=Any):
    _ = conn
    q = "pragma table_info('info')"
    cursor.execute(q)
    ret = cursor.fetchall()
    if len(ret) > 0:
        return False
    else:
        return True

@db_func
def get_summary_module_store_list(args={}, conn=None, cursor=None):
    if not conn or not cursor:
        return
    q = "select name, store from summary where publish_time >= ?"
    cursor.execute(q, (args.get("publish_time"),))
    ret = cursor.fetchall()
    out = []
    for r in ret:
        out.append({"name": r[0], "store": r[1]})
    return out


@db_func
def fetch_conf_cache(args={}, conn=None, cursor=None, conf={}):
    from requests import Session
    from .ov.account import get_current_id_token
    from ..system import get_cache_dir
    from .ov import get_store_url
    from os.path import join
    from ..util.util import quiet_print

    if not conn or not cursor:
        return
    module_stores = get_summary_module_store_list(args=args)
    if not module_stores:
        return
    id_token = get_current_id_token(args=args)
    params = {"idToken": id_token, "publish_time": args.get("publish_time")}
    s = Session()
    s.headers["User-Agent"] = "oakvar"
    quiet_print(f"fetching store cache 5/5...", args=args)
    for module_store in module_stores:
        name = module_store["name"]
        store = module_store["store"]
        fpath = join(get_cache_dir("conf", conf=conf), store, name + ".json")
        url = f"{get_store_url()}/fetch_conf/{store}/{name}"
        res = s.post(url, data=params)
        content = b"{}"
        if res.status_code == 200:
            content = res.content
        elif res.status_code == 404:
            content = b"{}"
        else:
            continue
        with open(fpath, "wb") as wf:
            wf.write(content)


@db_func
def fetch_logo_cache(args={}, conn=None, cursor=None, conf={}):
    from requests import Session
    from .ov.account import get_current_id_token
    from ..system import get_cache_dir
    from .ov import get_store_url
    from os.path import join
    from ..util.util import quiet_print

    if not conn or not cursor:
        return
    module_stores = get_summary_module_store_list(args=args)
    if not module_stores:
        return
    id_token = get_current_id_token(args=args)
    params = {"idToken": id_token, "publish_time": args.get("publish_time")}
    s = Session()
    s.headers["User-Agent"] = "oakvar"
    quiet_print(f"fetching store cache 4/5...", args=args)
    for module_store in module_stores:
        name = module_store["name"]
        store = module_store["store"]
        fpath = join(get_cache_dir("logo", conf=conf), store, name + ".png")
        url = f"{get_store_url()}/fetch_logo/{store}/{name}"
        res = s.post(url, data=params)
        content = b""
        if res.status_code == 200:
            content = res.content
        elif res.status_code == 404:
            content = b""
        else:
            continue
        with open(fpath, "wb") as wf:
            wf.write(content)


@db_func
def fetch_readme_cache(args={}, conn=None, cursor=None, conf={}):
    from requests import Session
    from .ov.account import get_current_id_token
    from ..system import get_cache_dir
    from .ov import get_store_url
    from os.path import join
    from ..util.util import quiet_print

    if not conn or not cursor:
        return
    module_stores = get_summary_module_store_list(args=args)
    if not module_stores:
        return
    id_token = get_current_id_token(args=args)
    params = {"idToken": id_token, "publish_time": args.get("publish_time")}
    s = Session()
    s.headers["User-Agent"] = "oakvar"
    quiet_print(f"fetching store cache 3/5...", args=args)
    for module_store in module_stores:
        name = module_store["name"]
        store = module_store["store"]
        fpath = join(get_cache_dir("readme", conf=conf), store, name)
        url = f"{get_store_url()}/fetch_readme/{store}/{name}"
        res = s.post(url, data=params)
        content = b""
        if res.status_code == 200:
            content = res.content
        elif res.status_code == 404:
            content = b""
        else:
            continue
        with open(fpath, "wb") as wf:
            wf.write(content)


@db_func
def fetch_summary_cache(args={}, conn=Any, cursor=Any):
    from requests import Session
    from .ov.account import get_current_id_token
    from ..exceptions import StoreServerError
    from ..exceptions import AuthorizationError
    from .ov import get_store_url
    from ..util.util import quiet_print

    _ = conn
    url = f"{get_store_url()}/fetch_summary"
    id_token = get_current_id_token(args=args)
    params = {"idToken": id_token, "publish_time": args.get("publish_time")}
    s = Session()
    s.headers["User-Agent"] = "oakvar"
    quiet_print(f"fetching store cache 1/5...", args=args)
    res = s.post(url, data=params)
    if res.status_code != 200:
        if res.status_code == 401:
            raise AuthorizationError()
        elif res.status_code == 500:
            raise StoreServerError()
        return False
    if args.get("clean_cache_db"):
        q = f"delete from summary"
        cursor.execute(q)
        conn.commit()
    res = res.json()
    cols = res["cols"]
    for row in res["data"]:
        q = f"insert or replace into summary ( {', '.join(cols)} ) values ( {', '.join(['?'] * len(cols))} )"
        cursor.execute(q, row)
    conn.commit()


@db_func
def fetch_versions_cache(args={}, conn=None, cursor=None):
    from requests import Session
    from json import dumps
    from .ov.account import get_current_id_token
    from ..exceptions import StoreServerError
    from ..exceptions import AuthorizationError
    from .ov import get_store_url
    from ..util.util import quiet_print
    from .consts import versions_table_cols

    if not conn or not cursor:
        return
    url = f"{get_store_url()}/fetch_versions"
    id_token = get_current_id_token(args=args)
    cols = dumps(versions_table_cols)
    params = {"idToken": id_token, "publish_time": args.get("publish_time"), "cols": cols}
    s = Session()
    s.headers["User-Agent"] = "oakvar"
    quiet_print(f"fetching store cache 2/5...", args=args)
    res = s.post(url, data=params)
    if res.status_code != 200:
        if res.status_code == 401:
            raise AuthorizationError()
        elif res.status_code == 500:
            raise StoreServerError(text=res.text)
        return False
    if args.get("clean_cache_db"):
        q = f"delete from versions"
        cursor.execute(q)
        conn.commit()
    res = res.json()
    cols = res["cols"]
    for row in res["data"]:
        q = f"insert or replace into versions ( {', '.join(cols)} ) values ( {', '.join(['?'] * len(cols))} )"
        cursor.execute(q, row)
    conn.commit()


@db_func
def get_local_last_updated(conn=None, cursor=None) -> str:
    from .consts import ov_store_last_updated_col

    if not conn or not cursor:
        return ""
    q = "select value from info where key=?"
    cursor.execute(q, (ov_store_last_updated_col,))
    res = cursor.fetchone()
    if not res:
        return ""
    last_updated = res[0]
    return last_updated


@db_func
def get_manifest(conn=None, cursor=None) -> Optional[dict]:
    import sqlite3

    if not conn or not cursor:
        return None
    cursor.row_factory = sqlite3.Row
    q = f"select distinct(name) from summary"
    cursor.execute(q)
    res = cursor.fetchall()
    mi = {}
    for r in res:
        name = r[0]
        m = module_info(name)
        if m:
            mi[name] = m.to_dict()
    return mi


@db_func
def get_urls(module_name: str, code_version: str, args={}, conn=None, cursor=None):
    from requests import Session
    from .ov.account import get_current_id_token
    from ..exceptions import StoreServerError
    from ..exceptions import AuthorizationError
    from .ov import get_store_url

    if not conn or not cursor:
        return
    q = f"select store from versions where name=? and code_version=?"
    cursor.execute(q, (module_name, code_version))
    ret = cursor.fetchall()
    store = None
    for r in ret:
        if not store or r[0] == "ov":
            store = r[0]
    if not store:
        return None
    url = f"{get_store_url()}/urls/{module_name}/{code_version}"
    id_token = get_current_id_token(args=args)
    params = {"idToken": id_token}
    s = Session()
    s.headers["User-Agent"] = "oakvar"
    res = s.post(url, data=params)
    if res.status_code == 200:
        return res.json()
    else:
        if res.status_code == 401:
            raise AuthorizationError()
        elif res.status_code == 500:
            raise StoreServerError()
        return


@db_func
def set_server_last_updated(args={}, conn=None, cursor=None):
    from .consts import ov_store_last_updated_col
    from ..consts import publish_time_fmt
    from datetime import datetime

    if not conn or not cursor:
        return
    if args:
        pass
    dt = datetime.now().strftime(publish_time_fmt)
    q = f"insert or replace into info ( key, value ) values ( ?, ? )"
    cursor.execute(q, (ov_store_last_updated_col, dt))
    conn.commit()


@db_func
def table_has_entry(table: str, conn=Any, cursor=Any) -> bool:
    _ = conn or cursor
    q = f"select count(*) from {table}"
    cursor.execute(q)
    v = cursor.fetchone()[0]
    return v > 0


@db_func
def check_tables(args={}, conn=Any, cursor=Any) -> bool:
    from ..util.util import quiet_print

    _ = conn or cursor
    for table in ["summary", "versions", "info"]:
        if not table_exists(table) or not table_has_entry(table):
            quiet_print(f"store cache table {table} does not exist.", args=args)
            return False
    return True


@db_func
def module_is_in_store(module_name: str, conn=Any, cursor=Any) -> bool:
    _ = conn
    q = f"select name from summary where name=?"
    cursor.execute(q, (module_name,))
    ret = cursor.fetchone()
    if ret:
        return True
    else:
        return False



def download(url, fpath, install_state=None, check_install_kill=None):
    from .download_library import download as download_util
    import gdown

    if "drive.google.com" in url:
        gdown.download(url=url, output=fpath, quiet=True, fuzzy=True)
    elif "github.com" in url:
        download_git_folder(url=url, install_dir=fpath)
    else:
        download_util(url, fpath, kind="file", verbose=False, replace=True, install_state=install_state, check_install_kill=check_install_kill)


def download_git_file(el, folder):
    import download as download_util
    from re import search

    if search(r"\.zip[0-9]*$", el["download_url"]):
        download_util.download(
            el["download_url"], str(folder), kind="file", verbose=False, replace=True
        )
    else:
        download_util.download(
            el["download_url"],
            str(folder / el["name"]),
            kind="file",
            verbose=False,
            replace=True,
        )


def get_git_api_url(url):
    from re import compile

    branch_re = compile("/(tree|blob)/(.+?)/")
    branch_match = branch_re.search(url)
    if not branch_match:
        return None
    url_1 = url[: branch_match.start()].replace("github.com", "api.github.com/repos", 1)
    url_2 = "contents"
    url_3 = url[branch_match.end() :]
    branch = branch_match.group(2)
    api_url = f"{url_1}/{url_2}/{url_3}?ref={branch}"
    return api_url


def download_git_folder(url=None, install_dir=None):
    from requests import get
    from pathlib import Path

    if not url or not install_dir:
        return
    if not isinstance(install_dir, Path):
        install_dir = Path(install_dir)
    api_url = get_git_api_url(url)
    if not api_url:
        return
    res = get(api_url)
    data = res.json()
    if isinstance(data, dict) and data.get("type") == "file":
        download_git_file(data, install_dir)
        return
    folder = install_dir / Path(url).stem
    if not folder.exists():
        folder.mkdir(parents=True, exist_ok=True)
    for el in data:
        if el["type"] == "dir":
            download_git_folder(url=el["html_url"], install_dir=folder)
        elif el["type"] == "file":
            download_git_file(el, folder)


def is_url(url):
    from re import compile

    url_pattern = compile("^(http|https)://.*$")
    match = url_pattern.match(url)
    return match is not None


def is_zip_path(url):
    from pathlib import Path

    p = Path(url)
    return p.exists() and p.suffix == ".zip"
